from osbot_gsuite.gsuite.drive.GDrive import GDrive
from osbot_utils.base_classes.Type_Safe import Type_Safe

from osbot_utils.utils import Misc


class GSlides(Type_Safe):
    presentations : object
    gdrive        : GDrive

    # def __init__(self, gsuite_secret_id=None):
    #     self.presentations = GSuite(gsuite_secret_id).slides_v1().presentations()
    #     self.gdrive        = GDrive(gsuite_secret_id)

    # misc utils

    def batch_update(self, file_id, requests):
        body = {'requests': requests}
        return self.execute(self.presentations.batchUpdate(presentationId=file_id, body=body))

    def execute(self,command):
        return self.gdrive.execute(command)

    def execute_requests(self, file_id, requests):
        return self.batch_update(file_id, requests)


    def random_id(self, prefix):
        return Misc.random_string_and_numbers(6, prefix + "_")


    def all_presentations(self):
        mime_type_presentations = 'application/vnd.google-apps.presentation'
        return self.gdrive.find_by_mime_type(mime_type_presentations)


    # Elements
    def element_create_image_request(self, page_id, image_url, x_pos=200, y_pos=200, width=100, height=100):
        return {  "createImage": {
                  "url"        : image_url,
                  "elementProperties": {
                      "pageObjectId": page_id,
                      "size": { "width" : { "magnitude": width, "unit": "PT" },
                                "height": { "magnitude": height,"unit": "PT" }},
                      "transform": { "scaleX": 1, "scaleY": 1, "translateX": x_pos, "translateY": y_pos, "unit": "PT" }}}}

    def element_create_image(self, file_id, page_id, image_url, x_pos=200, y_pos=200, width=100, height=100):
        requests = [ self.element_create_image_request(page_id, image_url, x_pos, y_pos, width, height)]
        result = self.batch_update(file_id, requests)
        if result:
            result_replies = result.get('replies')
            if result_replies:
                return result_replies[0].get('createImage').get('objectId')

    def element_create_table_request(self, slide_id, rows=3, cols=3, x_pos=200, y_pos=200, width=100, height=100, objectId=None):
        return { "createTable": { "objectId"         : objectId,
                                  "elementProperties": { "pageObjectId": slide_id,
                                                         "size"        : {  "width": {"magnitude":  width, "unit": "PT"},
                                                                            "height": {"magnitude": height,"unit": "PT"}},
                                                         "transform"   : { "scaleX": 1, "scaleY": 1, "translateX": x_pos, "translateY": y_pos,"unit": "PT" }},
                                  "rows"             : rows                         ,
                                  "columns"          : cols                        }}
                                  #"tableRows"        : [{'rowHeight':{ 'magnitude': 10, 'unit': 'PT'}},{'rowHeight':{ 'magnitude': 100, 'unit': 'PT'}}]}}
    def element_create_table(self, file_id, slide_id, rows=3, cols=3, x_pos=200, y_pos=200, width=100, height=100, objectId=None):
        requests = [  self.element_create_table_request(slide_id, rows, cols, x_pos, y_pos, width, height, objectId) ]
        result = self.batch_update(file_id, requests)
        if result:
            return result.get('replies')[0].get('createTable').get('objectId')

    def element_create_shape_request(self, page_id, x_pos=200, y_pos=200, width=100, height=100, objectId=None):
        return { 'createShape': { 'objectId': objectId,
                                  'shapeType': 'TEXT_BOX',
                                  'elementProperties': {
                                      'pageObjectId': page_id,
                                      'size'        : { 'height': { 'magnitude': height, 'unit': 'PT'},
                                                        'width' : { 'magnitude': width , 'unit': 'PT'}},
                                      'transform'   : { 'scaleX': 1, 'scaleY': 1, 'translateX': x_pos, 'translateY': y_pos, 'unit': 'PT' }}}}

    def element_create_text_requests(self, page_id, text = "Text...", x_pos=200, y_pos=200, width=100, height=100, objectId=None):
        return [ self.element_create_shape_request(page_id, x_pos, y_pos, width, height, objectId),
                 self.element_insert_text_request(objectId,text)]

    def element_create_text(self,file_id, page_id, text = "Text...", x_pos=200, y_pos=200, width=100, height=100, objectId=None):
        requests = self.element_create_text_requests(page_id, text, x_pos, y_pos, width, height, objectId)
        result = self.batch_update(file_id, requests)
        if result:
            return result.get('replies')[0].get('createShape').get('objectId')

    def element_create_shape(self,file_id, page_id, shape_type, x_pos=200, y_pos=200, width=100, height=100):
        requests = [ { 'createShape': { 'shapeType': shape_type,
                                        'elementProperties': {
                                            'pageObjectId': page_id,
                                            'size'        : { 'height': { 'magnitude': height, 'unit': 'PT'},
                                                              'width' : { 'magnitude': width , 'unit': 'PT'}},
                                            'transform'   : { 'scaleX': 1, 'scaleY': 1, 'translateX': x_pos, 'translateY': y_pos, 'unit': 'PT' }}}}]

        result = self.batch_update(file_id, requests)
        if result:
            return result.get('replies')[0].get('createShape').get('objectId')

    def element_delete(self, file_id, element_id):
        requests = [ {  'deleteObject' : { 'objectId': element_id } } ]
        return self.batch_update(file_id, requests)

    def element_insert_text_request(self, objectId, text):
        return { 'insertText': { 'objectId': objectId, 'insertionIndex': 0, 'text': text }}

    def element_set_table_cell_size_bold_requests(self, table_id, row, col, size, bold):
        style = {"bold": bold, "fontSize": { "magnitude": size, "unit": "PT" }}
        fields = "bold,fontSize"
        return self.element_set_table_text_style_request(table_id, row,col,style,fields)

    def element_set_table_text_style_request(self, shape_id, row, col,style, fields):
        return {'updateTextStyle': { 'objectId'    : shape_id  ,
                                     "cellLocation": {"rowIndex": row, "columnIndex": col},
                                     'style'       : style     ,
                                     'fields'      : fields   }}

    def element_set_table_text_requests(self, table_id, row, col, text):
        return     [#{ "deleteText": {   "objectId"      : table_id,
                    #                    "cellLocation"  : {  "rowIndex": row, "columnIndex": col   },
                    #                    "textRange"     : {"type": "ALL"                         }}},
                    { "insertText": {   "objectId"      : table_id,
                                        "cellLocation"  : {  "rowIndex": row, "columnIndex": col   },
                                        "text"          : text,
                                        "insertionIndex": 0                                       }}]

    def element_set_table_text(self, file_id, table_id, row, col, text):
        requests = self.element_set_table_text_requests(table_id, row, col, text)

        self.execute_requests(file_id,requests)


    def element_set_table_column_width_request(self,table_id, column_index, column_width):

        return { 'updateTableColumnProperties': { 'objectId': table_id, "columnIndices": [column_index],
                                                  "tableColumnProperties": { 'columnWidth': { "magnitude": column_width, "unit": "PT" } },
                                                  "fields": "columnWidth" } }

    def element_set_table_cell_aligment_request(self, table_id, row_index, column_index, row_span, column_span, alignment='MIDDLE'):
        return { "updateTableCellProperties": { "objectId": table_id,
                                                "tableRange": {
                                                    "location": { "rowIndex":  row_index,"columnIndex": column_index },
                                                                  "rowSpan": row_span, "columnSpan": column_span },
                                                "tableCellProperties": { "contentAlignment": alignment },
                                                "fields": "contentAlignment"}}

    def element_set_table_row_height_request(self, table_id, height):
        return  { "updateTableRowProperties": {   "objectId": table_id,
                                                  "rowIndices": 0,
                                                  "tableRowProperties": { "minRowHeight":  { 'magnitude': height, 'unit': 'PT'}},
                                                  "fields"            : "minRowHeight"}}

    def element_set_text_requests(self, file_id, element_id, text):
        return [ {   'deleteText' : { 'objectId'      : element_id         ,
                                      'textRange'     : { 'type': 'ALL' }}},
                 {
                     'insertText': { 'objectId'      : element_id          ,
                                     'insertionIndex': 0                   ,
                                     'text'          : text              }}]
    def element_set_text(self, file_id, element_id, text):

        requests =  self.element_set_text_requests(file_id, element_id, text)

        return self.batch_update(file_id, requests)

    def element_set_text_style_requests(self, object_id, style, fields):
        return {'updateTextStyle': { 'objectId': object_id  ,
                                     'style'   : style     ,
                                     'fields'  : fields   }}

    def element_set_text_style(self, file_id, shape_id,style, fields):
        requests = [ self.element_set_text_style_requests(shape_id, style, fields) ]
        return self.batch_update(file_id, requests)


    def element_set_text_style_requests__for_title(self, shape_id, font_size, blue=0.5, green=0.5, red=0.5):
        style = {  'bold'            : True,
                   'fontFamily'      : 'Avenir',
                   'fontSize'        : { 'magnitude'  : font_size, 'unit': 'PT' },
                   'foregroundColor' : {'opaqueColor': {'rgbColor': {'blue': blue, 'green': green, 'red': red}}}}
        fields   = 'bold,fontFamily,fontSize,foregroundColor'


        return self.element_set_text_style_requests(shape_id, style, fields)

    def element_set_shape_properties(self, file_id, shape_id,properties , fields=None):
        if fields is None:
            fields = ",".join(list(set(properties)))
        requests = [{'updateShapeProperties': { 'objectId'         : shape_id     ,
                                                'shapeProperties'  : properties   ,
                                                'fields'           : fields     }}]
        return self.batch_update(file_id, requests)

    def presentation_create(self, title):
        body = { 'title': title }
        presentation = self.presentations.create(body=body).execute()
        return presentation.get('presentationId')

    def presentation_copy(self, file_id, title, parent_folder):
        body  = { 'name': title, 'parents':[parent_folder] }
        result = self.execute(self.gdrive.files.copy(fileId = file_id, body=body))
        return result.get('id')

    def presentation_metadata(self,presentation_id):
        try:
            return self.presentations.get(presentationId = presentation_id).execute()
        except:
            return None

    def slide_delete_request(self, slide_id):
        return { "deleteObject": { "objectId" : slide_id}}

    def slide_delete(self, presentation_id, slide_id):
        requests =   [ self.slide_delete_request(slide_id) ]
        return self.execute_requests(presentation_id, requests)

    def slide_copy(self,presentation_id, slide_id, new_slide_id, objects_ids = {}):
        requests =   [ { "duplicateObject": {
                                                "objectId" : slide_id,
                                                "objectIds": objects_ids}} ]
        requests[0]['duplicateObject']['objectIds'][slide_id] = new_slide_id
        return self.execute_requests(presentation_id, requests)

    def slide_create_request(self, new_slide_id=None, layout='BLANK' , insert_at= None ):
        return { "createSlide": {       "objectId"      : new_slide_id,
                                        'insertionIndex': insert_at,
                                        'slideLayoutReference': { 'predefinedLayout': layout } }}

    def slide_create(self, presentation_id, insert_at=1, layout='TITLE', new_slide_id=None):
        requests =   [ self.slide_create_request(new_slide_id, layout, insert_at)]
        result = self.execute_requests(presentation_id, requests)
        if result:
            return result.get('replies')[0].get('createSlide').get('objectId')


    def slide_move_to_pos_request(self, presentation_id, slide_id, pos):
        return [{ "updateSlidesPosition": { "slideObjectIds": [ slide_id],
                                            "insertionIndex": pos }}]
    def slide_move_to_pos(self, presentation_id, slide_id, pos):
        requests = self.slide_move_to_pos_request(presentation_id, slide_id, pos)
        return self.execute_requests(presentation_id,requests)

    def slide_elements(self, presentation_id, page_number):
        slides = self.slides(presentation_id)
        page   = slides[page_number]
        if page:
            return page.get('pageElements')
        return []

    def slide_elements_via_id(self, presentation_id, slide_id):
        slides = self.slides_indexed_by_id(presentation_id)
        slide = slides.get(slide_id)
        if slide:
            return slide.get('pageElements')
        return []

    def slide_elements_via_id_indexed_by_id(self, presentation_id, slide_id):
        elements = {}
        for element in self.slide_elements_via_id(presentation_id,slide_id):
            elements[element.get('objectId')] = element
        return elements

    def slides(self, presentation_id):
        presentation = self.presentation_metadata(presentation_id)
        if presentation:
            return presentation.get('slides')
        return []

    def slides_indexed_by_id(self, presentation_id):
        presentation = self.presentation_metadata(presentation_id)
        slides = {}
        if presentation:
            for slide in presentation.get('slides'):
                slides[slide.get('objectId')] = slide
        return slides



    # Helper methods

    def add_slide_with_table_from_array(self, file_id, slide_id, title, data, row_widths = []):
        title_id  = '{0}_title'.format(slide_id)
        table_id  = '{0}_table'.format(slide_id)
        headers   = data.pop(0)
        cols      = len(headers)
        rows      = len(data) + 1
        cell_size = 7
        self.slide_delete(file_id, slide_id)
        requests = [ self.slide_create_request                      (slide_id                           ),
                     self.element_create_shape_request              (slide_id, 10, 10, 500, 50, title_id),
                     self.element_insert_text_request               (title_id, title),
                     self.element_set_text_style_requests__for_title(title_id, 26),
                     self.element_create_table_request              (slide_id, rows, cols, 12, 80, 700, 270, table_id),
                     #self.element_set_table_row_height_request      (table_id, 10)
                    ]

        for col_index, header in enumerate(headers):
            requests.extend(self.element_set_table_text_requests          (table_id, 0,col_index, header  ))
            requests.append(self.element_set_table_cell_size_bold_requests(table_id, 0,col_index, 11, True))

        for row_index, row in enumerate(data):
            for col_index, cell in enumerate(row):
                requests.extend(self.element_set_table_text_requests(table_id          , row_index + 1, col_index, cell))
                requests.append(self.element_set_table_cell_size_bold_requests(table_id, row_index + 1, col_index, cell_size, False))

        for index, row_width in enumerate(row_widths):
            requests.append(self.element_set_table_column_width_request(table_id, index, row_width))

        requests.append(self.element_set_table_cell_aligment_request(table_id, 0,0, rows,cols))

        self.execute_requests(file_id, requests)
        return table_id

    def add_slide_with_table_from_object(self, file_id, slide_id, title, data):

        title_id = '{0}_title'.format(slide_id)
        table_id = '{0}_table'.format(slide_id)
        headers  = list(set(data))
        rows     = len(headers)                        # number of fields to show
        cols     = 2                                   # name:value pair
        requests = [ self.slide_create_request(slide_id),
                     self.element_create_shape_request              (slide_id, 10, 10, 500,  50, title_id                ),
                     self.element_insert_text_request               (title_id,title                                      ),
                     self.element_set_text_style_requests__for_title(title_id, 26                                        ),
                     self.element_create_table_request              (slide_id, rows , cols, 12, 80, 700, 270, table_id   ),
                     self.element_set_table_column_width_request    (table_id, 0 ,110                                    ),
                     self.element_set_table_column_width_request    (table_id, 1, 585                                    )]

        for index, header in enumerate(headers):
           requests.extend(self.element_set_table_text_requests          (table_id, index, 0, header))
           requests.extend(self.element_set_table_text_requests          (table_id, index, 1, str(data[header])))
           requests.append(self.element_set_table_cell_size_bold_requests(table_id, index, 0, 10, True))
           requests.append(self.element_set_table_cell_size_bold_requests(table_id, index, 1, 9 , False))

        #requests.extend(self.element_set_table_text_requests        (table_id, 1, 0, headers.pop()))

        self.execute_requests(file_id, requests)

        #requests = [self.element_set_table_text_requests(table_id, 1, 1, 'headers.pop()')]

        #self.execute_requests(file_id, requests)


from time import time

from osbot_aws.helpers.Lambda_Helpers import slack_message
from osbot_gsuite import version_osbot_gsuite
from osbot_gsuite.apis.create_slides.Project_Slides              import Project_Slides
from osbot_gsuite.apis.create_slides.Slides_for_Projects         import Slides_for_Projects

class Slides_Commands:

    @staticmethod
    def _gs_functions_ids():
        slides_for_projects = Slides_for_Projects()
        key = 'IA-402'
        issue = list(slides_for_projects.elastic().search_on_field_for_value('Key.keyword', key)).pop(0)
        return  issue.get('Issue Links').get('is parent of')

    @staticmethod
    def gs_functions(team_id, channel, params):
        slides_for_projects = Slides_for_Projects()

        functions = list(slides_for_projects.elastic().search_on_field_for_values('Key.keyword', Slides_Commands._gs_functions_ids()))
        attachment_text = ''
        for function in functions:
            key     = function.get('Key')
            summary = function.get('Summary')
            link    = 'https://jira.photobox.com/browse/{0}'.format(key)
            attachment_text += ' <{0}|{1}>  - {2} \n'.format(link, summary, key)
        text = ':point_right: Here is the current list of GS Functions (use `slides create {{jira id}}` to create the slides:) '
        attachments = [{'text': attachment_text}]
        return text, attachments

    @staticmethod
    def gs_projects(team_id, channel, params):
        slides_for_projects = Slides_for_Projects()
        #if len(params) ==0 :

        projects = slides_for_projects.gs_projects_by_title()
        text        = ':point_right: Here is the current list of `{0}` GS Projects (use `slides create {{jira id}}` to create the slides:) '.format(len(projects))
        attachment_text=''

        for index, title in enumerate(sorted(projects.keys())):
            project = projects[title]
            key   = project.get('Key')
            attachment_text += '{0}. {1}  - {2} \n'.format(index + 1,title, key)  # can't use the link here since it will get top big and slack will not show all results
        attachments = [{'text': attachment_text }]
        return text,attachments

    @staticmethod
    def create(team_id, channel, params):
        try:
            jira_key = params.pop(0)
            options  = ''
            if len(params) > 0: options  = params.pop(0)
            # jira_link = 'https://jira.photobox.com/browse/{0}'.format(jira_key)
            # slack_message(":zero: Creating slides for project: `{0}` (which you can edit in jira :point_right: <{1}|here>)".format(jira_key,jira_link), [],channel, team_id)jira_key = params.pop(0)

            jira_link = 'https://jira.photobox.com/browse/{0}'.format(jira_key)
            slack_message(":one: Creating slides for: `{0}` (which you can edit in jira :point_right: <{1}|here>)".format(jira_key,jira_link), [],channel, team_id)
            project_slides = Project_Slides()
            start = time()

            if jira_key in Slides_Commands._gs_functions_ids(): # is an GS Function
                slack_message(":two: Creating slides for Function `{0}`".format(jira_key), [],channel, team_id)
                weblink = project_slides.create_slides_for_function(jira_key)
            else:
                weblink = project_slides.create_slides_text_content(jira_key)
                if options == 'no_graphs':
                    slack_message(":one: ...skipping graph creation...", [],channel, team_id)
                else:
                    slack_message(":two: Currently adding the graphs for project `{0}`, but if you want, click <{1}|here to open the Slides>".format(jira_key,weblink), [],channel, team_id)
                    project_slides.create_slides_graphs()           \
                                  .create_slides_appendix()
            duration = time()-start
            return ":three: Slide creation complete. You can see it at <{2}|{0} Project slides>  \n :point_right: _creation duration: `{1:.0f}` secs_\n :point_right: _create pdf using: `gdocs pdf {3}`_".format(jira_key, duration,weblink, project_slides.file_id),[]
        except Exception as error:
            slack_message("Error presentation for project: `{0}`: {1}".format(jira_key,error), [], channel, team_id)
        return None,None

    @staticmethod
    def gs_services(team_id, channel, params):
        slides_for_projects = Slides_for_Projects()
        services = slides_for_projects.gs_services_by_title()
        text = ':point_right: Here is the current list of `{0}` GS Services: '.format(len(services))
        attachment_text = ''

        for index, title in enumerate(sorted(services.keys())):
            service = services[title]
            key  = service.get('Key')
            link  = 'https://jira.photobox.com/browse/{0}'.format(key)
            attachment_text += ' {0}.  <{1}|{2}>  - {3} \n'.format(index + 1 , link, title, key)
        attachments = [{'text': attachment_text}]
        return text, attachments


    @staticmethod
    def create_projects(team_id, channel, params):
        key = 'GSSP-112'            # FY20 projects
        Project_Slides().create_project_slides_for_functions(key,team_id,channel)

    @staticmethod
    def version(team_id, channel, params):
        return version_osbot_gsuite



from googleapiclient.discovery          import build
from google.oauth2.credentials          import Credentials
from googleapiclient.discovery          import Resource
from osbot_utils.utils.Json import json_loads

from osbot_utils.base_classes.Type_Safe import Type_Safe
from osbot_utils.utils.Env              import get_env

ENV_NAME__GSUITE__OAUTH2__DATA = 'GSUITE__OAUTH2__DATA'
ENV_NAME__GSUITE__OAUTH2__FILE = 'GSUITE__OAUTH2__FILE'


class GSuite(Type_Safe):                                           # todo see if there is a better name for this

    def gsuite__oauth2__data(self):
        return get_env(ENV_NAME__GSUITE__OAUTH2__DATA)

        # return file with token credentials
    def gsuite__oauth2__file(self):
        return get_env(ENV_NAME__GSUITE__OAUTH2__FILE)


    # this creates the credentials object required to create the GSuite service object
    def create_credentials(self, scopes=None) -> Credentials:
        oauth2_data = self.gsuite__oauth2__data()
        if oauth2_data:
            info        = json_loads(oauth2_data)
            credentials = Credentials.from_authorized_user_info(info, scopes=scopes)
            return credentials
        oauth2_file = self.gsuite__oauth2__file()
        if oauth2_file:
            credentials = Credentials.from_authorized_user_file(filename=oauth2_file, scopes=scopes)
            return credentials

        raise ValueError("no OAuth2 data or file found")


    def create_service(self,service_name, version, scopes) -> Resource:
        creds = self.create_credentials(scopes)
        return build(service_name, version, credentials=creds)

    # helper files to create individual GSuite service objects
    def admin_reports_v1(self):
        return self.create_service('admin', 'reports_v1','admin.reports.audit.readonly')

    def calendar_v3(self):
        return self.create_service('calendar','v3','calendar')

    def docs_v1(self):
        return self.create_service('docs', 'v1', 'documents')

    def drive_v3(self):
        kwargs = dict(service_name = 'drive',
                      version      = 'v3',
                      scopes       = ['https://www.googleapis.com/auth/drive'])
        return self.create_service(**kwargs)

    def drive_activity_v2(self):
        return self.create_service('driveactivity', 'v2', 'drive.activity')

    def people_v1(self):
        return self.create_service('people', 'v1', 'contacts')

    def slides_v1(self):
        return self.create_service('slides', 'v1', 'presentations')

    def sheets_v4(self):
        return self.create_service('sheets', 'v4', 'spreadsheets')




from osbot_elastic.Elastic_Search import Elastic_Search

from osbot_gsuite.apis.GDrive                           import GDrive
from osbot_gsuite.apis.GSheets                          import GSheets
from osbot_gsuite.apis.GSlides                          import GSlides
from osbot_gsuite.apis.create_slides.GSBot_Helpers      import GSBot_Helper
from osbot_gsuite.apis.create_slides.GSBot_to_GDrive    import GSBot_to_GDrive



class Slides_for_Projects:
    def __init__(self, gsuite_secret_id=None):
        self.file_id                = '1mmDcM4JzrYvLjVd2werl1jXhEaCRlksCSRpocGRI0H4'
        self.index                  = 'it_assets,jira'
        self.secret_id              = 'elastic-jira-dev-2'
        self.gsuite_secret_id       = gsuite_secret_id
        self.master_slide           = 'g4cb8e4f3f7_0_0'
        self.master_title_id        = 'g4cb8e4f3f7_0_3'
        self.master_link_id         = 'g4eced7afc3_1_0'
        self.master_table_id        = 'g4cb8e4f3f7_0_2'
        self.master_graphs_slide_id = 'g4ecf9bad4f_0_0'
        self.master_graphs_title_id = 'g4ecf9bad4f_0_3'
        self.master_graphs_table_id = 'g4ecf9bad4f_0_2'
        self.jira_link              = 'https://jira.photobox.com/browse/{0}'
        self._elastic               = None
        self._gslides               = None
        self._gdrive                = None
        self._gsheets               = None

    def add_graph_to_slide(self, slide_id,graph_name_or_key, depth,links_path,x,y,width,height):
        gsuite_secret_id = 'gsuite_gsbot_user'          # to refactor
        gsbot_helper     = GSBot_Helper()
        gsbot_to_GDrive  = GSBot_to_GDrive(gsuite_secret_id)
        png_file         = gsbot_helper.get_png_from_expanded_graph(graph_name_or_key, depth,links_path)
        image_id         = gsbot_to_GDrive.upload_png_file_to_gdrive(png_file)

        image_url = "https://lh3.google.com/u/1/d/{0}".format(image_id)

        return self.gslides().element_create_image(self.file_id, slide_id, image_url, x, y, width, height)


    def create_slide(self, slide_to_copy_id, new_slide_id, object_ids = {}):
        #slides = self.gslides().slides_indexed_by_id(self.file_id)
        return self.gslides().slide_copy(self.file_id, slide_to_copy_id, new_slide_id,object_ids)

    def create_slide_with_one_graph(self,slide_id, title, graph_1):
        title_id = '{0}_title'.format(slide_id)
        table_id = '{0}_table'.format(slide_id)
        object_ids               = {
                                        self.master_graphs_title_id : title_id,
                                        self.master_graphs_table_id : table_id
                                   }
        self.gslides().slide_delete(self.file_id, slide_id)
        self.create_slide(self.master_graphs_slide_id, slide_id, object_ids)
        self.gslides().element_set_text(self.file_id, title_id, title)
        self.gslides().element_delete(self.file_id, table_id)

        self.add_graph_to_slide(slide_id, graph_1[0], graph_1[1], graph_1[2], 50, 60, 500, 300)

    def create_slide_with_two_graphs(self, slide_id, title, graph_1=None, graph_2=None):
        title_id                 = '{0}_title'.format(slide_id)
        table_id                 = '{0}_table'.format(slide_id)
        object_ids               = {
                                        self.master_graphs_title_id : title_id ,
                                        self.master_graphs_table_id : table_id
                                   }
        self.gslides().slide_delete(self.file_id,slide_id)
        self.create_slide(self.master_graphs_slide_id,slide_id, object_ids)
        #title                    = "Graphs - Risks and Stakeholders"
        self.gslides().element_set_text(self.file_id, title_id, title)
        #self.slides_for_projects.add_graph_to_slide(slide_id, 'GSSP-128','7','affects,stakeholders_up'      , 120, 200, 500, 200)
        if graph_1:
            self.gslides().element_set_table_text(self.file_id, table_id, 0,0, graph_1[0])
            self.add_graph_to_slide(slide_id, graph_1[1], graph_1[2], graph_1[3], 120, 60, 500, 130)
        if graph_2:
            self.gslides().element_set_table_text(self.file_id, table_id, 1, 0, graph_2[0])
            self.add_graph_to_slide(slide_id, graph_2[1], graph_2[2], graph_2[3], 130, 220, 500, 130)

    def elastic(self):
        if self._elastic is None:
            self._elastic = Elastic_Search()._setup_Elastic_on_cloud_via_AWS_Secret(self.index, self.secret_id)
        return self._elastic

    def get_slide(self, slide_id):
        #slides = self.gslides().slides_indexed_by_id(self.file_id)  # get all slides
        #slide  = slides.get(slide_id)                               # search for slide
        #if slide:                                                   # return if it exists already
        #    return slide
        self.gslides().slide_delete(self.file_id,slide_id)
        object_ids = {
                        self.master_title_id : '{0}_title'.format(slide_id),
                        self.master_link_id  : '{0}_link' .format(slide_id),
                        self.master_table_id : '{0}_table'.format(slide_id)
                        }
        self.create_slide(self.master_slide, slide_id,object_ids)   # create slide
        return self.gslides().slides_indexed_by_id(self.file_id)    # get slide data

    def gslides(self):
        if self._gslides is None:
            self._gslides = GSlides(self.gsuite_secret_id)
        return self._gslides

    def gdrive(self):
        if self._gdrive is None:
            self._gdrive = GDrive(self.gsuite_secret_id)
        return self._gdrive

    def gsheets(self):
        if self._gsheets is None:
            self._gsheets = GSheets(self.gsuite_secret_id)
        return self._gsheets


    def gs_project(self,title):
        result = list(self.elastic().search_on_field_for_value('Summary.keyword', title))
        if len(result) == 1:
            return result.pop()
        return None

    def gs_project_by_Id(self,key):
        result = list(self.elastic().search_on_field_for_value('Key.keyword', key))
        if len(result) == 1:
            return result.pop()
        return None


    def gs_projects(self):
        return list(self.elastic().search_on_field_for_value('Issue Type','Project'))

    def gs_projects_by_id(self):
        projects = {}
        for item in self.gs_projects():
            projects[item.get('Key')] = item
        return projects

    def gs_projects_by_title(self):
        projects = {}
        for item in self.gs_projects():
            projects[item.get('Summary')] = item
        return projects

    def gs_services(self):
        return list(self.elastic().search_on_field_for_value('Issue Type','"Service"'))

    def gs_services_by_id(self):
        services = {}
        for item in self.gs_services():
            services[item.get('Key')] = item
        return services

    def gs_services_by_title(self):
        services = {}
        for item in self.gs_services():
            services[item.get('Summary')] = item
        return services


    def set_slide_title_request(self, slide_id, title):
        title_id = '{0}_title'.format(slide_id)
        return self.gslides().element_set_text_requests(self.file_id, title_id, title)

    def set_slide_link_requests(self, slide_id, jira_key):
        link_id = '{0}_link'.format(slide_id)
        requests = self.gslides().element_set_text_requests(self.file_id, link_id, jira_key)
        requests.extend([{ 'updateTextStyle': { 'objectId': link_id,
                                                'style'   : { 'link': { 'url': self.jira_link.format(jira_key) } },
                            'fields': 'link'  }}])
        return requests

    def set_slide_link(self, slide_id, jira_key):
        return self.set_slide_link_requests(slide_id, jira_key)


    def set_table_text_request(self, slide_id, row, col, text):
        table_id = "{0}_table".format(slide_id)
        return self.gslides().element_set_table_text_requests(table_id, row, col, text)



    # main methods with business logic
    def create_slide_with_project_data(self, jira_key):
        project_data      = self.gs_project_by_Id(jira_key)
        slide_id          = 'project-{0}'.format(jira_key)
        title             = project_data.get('Summary')
        it_system         = str(project_data.get('Issue Links').get('affects'))
        risks             = str(project_data.get('Issue Links').get('fixes' ))
        gs_services       = str(project_data.get('Issue Links').get('uses'  ))
        description       = project_data.get('Description')
        pilar             = 'TBD'
        status            = project_data.get('Status')
        management_owner  = 'TBD'
        risk_owner        = 'TBD' #project_data.get('Risk Owner','TBD')
        vulns             = '...'
        budget            = 'TBD'
        business_services = 'TBD'

        self.get_slide(slide_id)
        requests = []
        requests.extend(self.set_slide_title_request(slide_id, title))
        requests.extend(self.set_slide_link (slide_id, jira_key))

        requests.extend(self.set_table_text_request (slide_id, 0, 0, it_system))
        requests.extend(self.set_table_text_request (slide_id, 0, 2, management_owner))
        requests.extend(self.set_table_text_request (slide_id, 0, 3, pilar))
        requests.extend(self.set_table_text_request (slide_id, 0, 4, risk_owner))
        requests.extend(self.set_table_text_request (slide_id, 1, 1, description))
        requests.extend(self.set_table_text_request (slide_id, 2, 1, risks))
        requests.extend(self.set_table_text_request (slide_id, 3, 1, vulns))
        requests.extend(self.set_table_text_request (slide_id, 4, 1, budget))
        requests.extend(self.set_table_text_request (slide_id, 5, 0, gs_services))
        requests.extend(self.set_table_text_request (slide_id, 5, 3, status))
        requests.extend(self.set_table_text_request (slide_id, 5, 2, business_services))
        self.gslides().execute_requests(self.file_id,requests)


    def create_slides_with_graphs(self, jira_key):
        self.create_slide_with_two_graphs('graph_{0}_gs_services_none'.format(jira_key)     ,
                                          '{0} - GS Services'.format(jira_key)              ,
                                          ('GS Services', jira_key , '3', 'uses,is child of'),
                                           None)
        self.create_slide_with_two_graphs('graph_{0}_risks_stakeholders'.format(jira_key)     ,
                                          '{0} - Risks and Stakeholders'.format(jira_key)     ,
                                          ('Risks'       , jira_key, '7', 'fixes,risks_up,creates VULN'),
                                          ('Stakeholders', jira_key, '7','affects,stakeholders_up'      ))


    def create_slides_with_projects_and_services(self, function_name, projects_key, services_key):
        #Dev.print("create_slides_with_projects_and_services: {0}  {1} {2}".format(function_name, projects_key, services_key))
        self.create_slide_with_one_graph('graphs_gs_projects',
                                         '{0} Projects'.format(function_name),
                                         (projects_key, '1', 'delivered by'))


        self.create_slide_with_one_graph('graphs_gs_services',
                                         '{0} Services'.format(function_name),
                                          (services_key, '1', 'is parent of'))

    def set_slides_positions(self,projects_ids):

        requests = []
        index = 1
        def set_position(slide_id):
            nonlocal index
            requests.extend(self.gslides().slide_move_to_pos_request(self.file_id, slide_id,index))
            index += 1


        set_position('graphs_gs_projects')
        set_position('graphs_gs_services')

        for jira_key in projects_ids:
            set_position('project-{0}'                  .format(jira_key))
            set_position('graph_{0}_gs_services_none'   .format(jira_key))
            set_position('graph_{0}_risks_stakeholders' .format(jira_key))

        set_position(self.master_slide)
        set_position(self.master_graphs_slide_id)
        self.gslides().execute_requests(self.file_id,requests)

    def export_to_pdf(self, function_name):
        target_file = '/tmp/gs_{0}_slides.pdf'.format(function_name)
        gdrive = self.gdrive()
        return gdrive.file_export_as_pdf_to(self.file_id, target_file)

    def create_slides_for_function___Cloud(self):
        projects_key = 'GSSP-116'
        services_key = 'IA-418'
        function_name = 'Cloud'
        self.create_slides_with_projects_and_services(function_name, projects_key, services_key)

        #service_ids  = self.elastic().get_data(services_key).get('_source').get('Issue Links').get('is parent of')
        projects_ids = self.elastic().get_data(projects_key).get('_source').get('Issue Links').get('delivered by')

        #projects_ids = projects_ids#[:4]

        for jira_key in projects_ids:
            self.create_slide_with_project_data(jira_key)
            self.create_slides_with_graphs(jira_key)

        self.set_slides_positions(projects_ids)







from osbot_gsuite.apis.GDrive           import GDrive
from osbot_utils.utils.Dev import Dev
from osbot_utils.utils.Files import Files


class GSBot_to_GDrive:
    def __init__(self,gsuite_secret_id=None):
        self.target_folder  = 'gsbot-graphs'
        self.gdrive         = GDrive(gsuite_secret_id)

    def target_folder_id(self):
        mime_type = 'application/vnd.google-apps.folder'
        return self.gdrive.find_by_name(self.target_folder, mime_type).get('id')

    def graph_id_in_gdrive(self, file_name):
        file_metadata = self.gdrive.find_by_name(file_name)
        if file_metadata:
            return file_metadata.get('id')
        return None

    def upload_png_file_to_gdrive(self, png_file):
        #png_file  = GSBot_Helper().get_png_from_saved_graph(graph_name)
        file_name = Files.file_name(png_file)
        folder_id = self.target_folder_id()
        file_id   = self.graph_id_in_gdrive(file_name)

        if file_id:         # if the file exists, delete it before uploading the new version
            Dev.pprint('deleting file: {0}'.format(file_id))
            self.gdrive.file_delete(file_id)
        return self.gdrive.file_upload(png_file, 'image/png', folder_id)
        # can't use the code below because, the update wasn't working due to GSlides keeping (somewhere) a cache of the previous value
        #                                   update: (Oct 2022): GSlides will create a copy of the file uploaded to googleusercontent.com (which is what is then used in the GSlides)
        # if file_id is None:
        #     return self.gdrive.file_upload(png_file,'image/png', folder_id)
        # else:
        #     return self.gdrive.file_update(png_file, 'image/png', file_id )



from osbot_aws.helpers.Lambda_Helpers import log_to_elk
from osbot_gsuite.apis.handlers.Lambda_Sheets import Lambda_Sheets


def run(event, context):
    try:
        from osbot_aws.apis.Lambda import load_dependency
        load_dependency("gmail")
        load_dependency("elastic")                                      # load dependency (download and unzip if first run)
        load_dependency("slack")

        Lambda_Sheets().handle_lambda_event(event)                      # invoke lambda handler from Lambda_Sheets class
        return "200 OK"

    except Exception as error:
        message = "[lambda_sheets] Error: {0}".format(error)
        log_to_elk(message, level='error')
        return message


import base64
import json

from osbot_aws.apis.Lambda import Lambda
from osbot_elastic.Elastic_Search import Elastic_Search
from osbot_utils.utils.Files import Files


class GSBot_Helper:

    def elastic__setup_elk(self):
        index           = 'save_to_elk'
        secrets_elastic = 'elastic-logs-server-1'
        return Elastic_Search()._setup_Elastic_on_cloud_via_AWS_Secret(index, secrets_elastic)

    def create_graph(self, start, direction, depth,view=None):
        lambda_graph = Lambda('pbx_gs_python_utils.lambdas.gs.elastic_jira')

        payload = { "params": ['links', start, direction, depth,view] }
        result = lambda_graph.invoke(payload)
        data = json.loads(result.get('text'))
        return data

    def expand_graph(self, graph_name, depth,links_path):
        lambda_graph = Lambda('lambdas.gsbot.gsbot_graph')

        payload = { 'data': {}, "params": ['expand', graph_name, depth, links_path]}
        result = lambda_graph.invoke(payload)
        return json.loads(result)

    def get_png_from_expanded_graph(self, graph_name, depth,links_path):
        target_file = '/tmp/puml_{0}_{1}_{2}.png'.format(graph_name, depth,links_path)
        #if Files.not_exists(target_file):
        graph_metadata = self.expand_graph(graph_name, depth,links_path)
        self.puml_to_png(graph_metadata.get('puml'), target_file)
        return target_file

    def find_R2s_for_jira_issue(self, jira_key):
        r2s = []
        graph_metadata = self.expand_graph(jira_key, 10, 'risks_up')
        for edge in graph_metadata.get('edges'):
            if edge[1] == 'creates R2':
                r2s.append(edge[2])
        return list(set(r2s))

    def get_png_from_new_graph(self, start, direction, depth, view=None):
        target_file = '/tmp/puml_{0}_{1}_{2}.png'.format(start,direction,depth)
        #if Files.not_exists(target_file):
        graph_metadata = self.create_graph(start,direction,depth,view)
        self.puml_to_png(graph_metadata.get('puml'), target_file)
        return target_file

    def get_png_from_saved_graph(self, graph_name, target_file=None):
        if target_file is None:
            target_file  = '/tmp/puml_{0}.png'.format(graph_name)
        else:
            Files.delete(target_file)
        if Files.not_exists(target_file):
            lucene_query = 'doc_data.name:"{0}"'           .format(graph_name)
            result       = self.elastic__setup_elk().search_using_lucene_index_by_id(lucene_query, 1, "date:desc").values()
            puml         =  list(result)[0].get('doc_data').get('extra_data').get('puml')
            self.puml_to_png(puml, target_file)

        return target_file

    def get_png_from_elk_dashboard(self, jira_key):
        png_file = '/tmp/puml_elk_dashboard_{0}.png'.format(jira_key)
        lambda_browser = Lambda('browser.lambda_browser')

        payload = {"params": ['elk', 'dashboard_project', jira_key]}
        png_data = lambda_browser.invoke(payload)
        with open(png_file, "wb") as fh:
            fh.write(base64.decodebytes(png_data.encode()))
        return png_file

    def get_png_from_browser_risks(self, risks):
        #png_file = '/tmp/puml_browser_risks_{0}.png'.format(jira_key)
        png_file = Files.temp_file(".png")
        lambda_browser = Lambda('lambdas.browser.lambda_browser')

        payload = {"params": ['risks', risks]}
        png_data = lambda_browser.invoke(payload)
        with open(png_file, "wb") as fh:
            fh.write(base64.decodebytes(png_data.encode()))
        return png_file



    def puml_to_png(self, puml, target_file):
        puml_to_png = Lambda('utils.puml_to_png')
        png_data = puml_to_png.invoke({"puml": puml})

        with open(target_file, "wb") as fh:
            fh.write(base64.decodebytes(png_data['png_base64'].encode()))



from osbot_gsuite.apis.GDrive import GDrive
from osbot_gsuite.apis.GSheet import GSheet
from osbot_gsuite.apis.GSuite import GSuite
from osbot_utils.decorators.lists.index_by import index_by
from osbot_utils.decorators.methods.cache_on_self import cache_on_self


class GSheets:

    def __init__(self, gsuite_secret_id=None):
        self.gdrive           = GDrive(gsuite_secret_id)
        self.gsuite_secret_id = gsuite_secret_id

    @cache_on_self
    def spreadsheets(self):
        return GSuite(self.gsuite_secret_id).sheets_v4().spreadsheets()

    def batch_update(self, file_id, requests):
        body = {'requests': requests}
        return self.execute(self.spreadsheets().batchUpdate(spreadsheetId=file_id, body=body))

    def create(self, title, folder=None):
        return self.gdrive.file_create("application/vnd.google-apps.spreadsheet", title, folder)

    def create_and_share_with_domain(self, title, domain, folder=None):
        file_id = self.gdrive.file_create("application/vnd.google-apps.spreadsheet", title, folder)
        self.gdrive.file_share_with_domain(file_id, domain)
        return file_id


        #return self.execute(self.spreadsheets().create( body=body))


    def execute(self,command):
        return self.gdrive.execute(command)

    def execute_request(self, file_id, request):
        return self.batch_update(file_id, [request])

    def execute_requests(self, file_id, requests):
        return self.batch_update(file_id, requests)


    # def all_spreadsheets(self):
    #    mime_type_presentations = 'application/vnd.google-apps.spreadsheet'
    #    return self.gdrive.find_by_mime_type(mime_type_presentations)

    def sheets_metadata(self, file_id):
        return self.execute(self.spreadsheets().get(spreadsheetId=file_id))

    @index_by
    def sheets(self, file_id):
        sheets = {}
        for sheet in self.sheets_metadata(file_id=file_id).get('sheets'):
            properties         = sheet.get('properties')
            sheet_id           = properties.get('sheetId')
            sheet_name         = properties.get('title'  )
            sheets[sheet_name] = self.sheet(file_id=file_id, sheet_id=sheet_id, sheet_name=sheet_name)
            #print(sheet)
        return sheets

    def sheet(self, file_id, sheet_id, sheet_name):
        if file_id is not None and sheet_id is not None and sheet_name is not None:
            return GSheet(gsheets=self, file_id=file_id, sheet_id=sheet_id, sheet_name=sheet_name)

    def sheets_add_sheet(self, file_id, title):
        request = { "addSheet": { "properties": { "title": title } } }

        result  =  self.execute_request(file_id, [request])
        return result.get('replies')[0].get('addSheet').get('properties').get('sheetId')


    def sheets_delete_sheet(self, file_id, sheet_id):
        request = { "deleteSheet": { "sheetId": sheet_id } }

        return self.execute_request(file_id, [request])

    def sheets_rename_sheet(self, file_id, sheet_id, new_name):
        request = {"updateSheetProperties": { "properties": { "sheetId": sheet_id    ,
                                                              "title"  : new_name   },
                                              "fields"    :   "title"               }}
        return self.execute_request(file_id, [request])

    def sheets_properties_by_id(self, file_id):
        values = {}
        metadata = self.sheets_metadata(file_id)
        for sheet in metadata.get('sheets'):
           properties = sheet.get('properties')
           sheet_id   = properties.get('sheetId')
           values[sheet_id] = properties
        return values

    def sheets_properties_by_title(self, file_id):
        values = {}
        metadata = self.sheets_metadata(file_id)
        if metadata:
            for sheet in metadata.get('sheets'):
               properties = sheet.get('properties')
               sheet_id   = properties.get('title')
               values[sheet_id] = properties
            return values

    def request_cell_set_background_color(self, sheet_id, col, row, red, green, blue):
        return {'updateCells': { 'start': {'sheetId': sheet_id, 'rowIndex': row, 'columnIndex': col },
                                 'rows': [{'values': [ {'userEnteredFormat' : {'backgroundColor': {'red': red, 'blue': blue, 'green': green}}}] } ],
                                 'fields': 'userEnteredFormat.backgroundColor'}}

    def request_cell_set_value(self, sheet_id, col, row, value):
        return {'updateCells': { 'start': {'sheetId': sheet_id, 'rowIndex': row, 'columnIndex': col },
                                 'rows': [{'values': [ {'userEnteredValue': {'stringValue': value}}] } ],
                                 'fields': 'userEnteredValue'}}

    def request_clear_formatting(self, sheet_id):
        return { "updateCells": { "range" : {"sheetId": sheet_id },
                                  "fields": "userEnteredFormat"  }}

    def clear_values(self, file_id, sheet_name):
        sheet_range = "{0}!A1:Z".format(sheet_name)
        return self.execute(self.spreadsheets().values().clear(spreadsheetId=file_id, range=sheet_range))

    def get_values(self, file_id, range):
        if file_id and range:
            values = self.spreadsheets().values()
            result = self.execute(values.get(spreadsheetId = file_id , range = range    ))
            return result.get('values', [])
        return []

    def set_sheet_data(self, file_id, sheet_name, data):
        sheet_data = self.covert_raw_data_to_flat_objects(data)
        self.clear_values(file_id, sheet_name)
        return self.set_values(file_id, sheet_name, sheet_data)

    def set_values(self, file_id, sheet_range, values):
        value_input_option = 'USER_ENTERED' # vs 'RAW'
        body               = { 'values' : values }
        result = self.execute(self.spreadsheets().values().update( spreadsheetId    = file_id,
                                                                 range            = sheet_range,
                                                                 valueInputOption = value_input_option,
                                                                 body             = body))
        return result

    # this assumes that the first row contains the fields and the rest is the data items
    def get_values_as_objects(self, file_id, range_selector):
        raw_data = self.get_values(file_id, range_selector)
        return self.convert_raw_data_to_objects(raw_data)

    def convert_raw_data_to_objects(self, raw_data):
        values = []
        if raw_data:
            fields = raw_data.pop(0)
            if not fields:                    # if the first row is empty, then use the second row as the fields
                fields = raw_data.pop(0)

            for column in raw_data:
                item = {}
                for index, cell in enumerate(column):
                    item[fields[index]] = cell
                values.append(item)
        return values


    # helper methods
    def covert_raw_data_to_flat_objects(self, raw_data):
        sheet_data = []
        headers = list(raw_data[0].keys())

        sheet_data.append(headers)

        for item in raw_data:
            sheet_data.append(list(item.values()))

        # make sure every cell is a string
        for i, row in enumerate(sheet_data):
            for j, cell in enumerate(row):
                if cell is None:
                    sheet_data[i][j] = ''
                else:
                    sheet_data[i][j] = str(sheet_data[i][j])
        return sheet_data

    def format_headers(self, file_id, sheet_id, end_column):
        requests =  [   { "repeatCell": { "range": {  "sheetId": sheet_id,
                                                      "startRowIndex"  : 0,
                                                      "endRowIndex"    : 1,
                                                      "endColumnIndex" : end_column},
                                          "cell" : {  "userEnteredFormat": { "backgroundColor": { "red": 0.8, "green": 0.8, "blue": 0.8 },
                                                                             "horizontalAlignment" : "CENTER",
                                                                             "textFormat"  : { "foregroundColor": { "red": 0.0, "green": 0.0, "blue": 0.0 },
                                                                                               "fontSize": 12,
                                                                                               "bold": True } }},
                                          "fields": "userEnteredFormat(backgroundColor,textFormat,horizontalAlignment)" } },
                        { "updateSheetProperties": { "properties": { "sheetId": sheet_id,
                                                                     "gridProperties": { "frozenRowCount": 1 }},
                                                     "fields": "gridProperties.frozenRowCount"}} ]
        self.execute_requests(file_id,requests)


from osbot_utils.utils import Misc

from osbot_gsuite.apis.create_slides.Slides_for_Projects import Slides_for_Projects


class Sheets_Commands:

    @staticmethod
    def _add_data(data, title):
        gsuite_secret_id = 'gsuite_gsbot_user'
        slides_for_projects = Slides_for_Projects(gsuite_secret_id)
        gsheets = slides_for_projects.gsheets()

        file_id    = '1ZIHavq-X_Ouo3ZIBn-zrwVTSU-PxWJ-Pt5zuNI66ahw'
        sheet_name    = title
        sheet_id      = gsheets.sheets_properties_by_title(file_id).get(sheet_name).get('sheetId')
        projects_data = []
        for project in data:
            item = {
                        'Key'        : project.get('Key') ,
                        'Title'      : project.get('Summary'),
                        'Description': Misc.remove_html_tags(project.get('Description')),
                        'Issue Links': project.get('Issue Links'),
                        'Rating'     : project.get('Rating'    ),
                        'Priority'   : project.get('Priority'  )}
            projects_data.append(item)
        sheet_data = gsheets.covert_raw_data_to_flat_objects(projects_data)

        gsheets.clear_values(file_id, sheet_name)
        gsheets.set_values(file_id, sheet_name,sheet_data)
        gsheets.format_headers(file_id, sheet_id, len(sheet_data[0]))

    @staticmethod
    def gs_projects(team_id, channel, params):
        gsuite_secret_id = 'gsuite_gsbot_user'
        slides_for_projects = Slides_for_Projects(gsuite_secret_id)

        gs_projects = slides_for_projects.gs_projects()
        Sheets_Commands._add_data(gs_projects, 'GS Projects')

        gs_services = slides_for_projects.gs_services()
        Sheets_Commands._add_data(gs_services, 'GS Services')

        gsheets = slides_for_projects.gsheets()
        file_id = '1ZIHavq-X_Ouo3ZIBn-zrwVTSU-PxWJ-Pt5zuNI66ahw'
        web_link = gsheets.gdrive.file_weblink(file_id)
        text = ":point_right: File updated ... you can see it at: {0} \n\n use `gdocs pdf {1}` to generate the pdf".format(web_link, file_id)
        return text,[]


    @staticmethod
    def live_demo(team_id, channel, params):
        return 42, []


from osbot_jira.api.jira_server.API_Jira import API_Jira
from osbot_jira.api.jira_server.API_Jira_Rest import API_Jira_Rest
from osbot_aws.helpers.Lambda_Helpers import slack_message
from osbot_gsuite.apis.GSheets import GSheets
from osbot_utils.utils.Dev import Dev


class API_Jira_Sheets_Sync:
    def __init__(self, file_id=None,gsuite_secret_id=None):
        self._gsheets           = None
        self._jira              = None
        self._jira_rest         = None
        #self._elastic           = None
        self._sheet_name        = None
        self._sheet_name_backup = None
        self._sheet_id          = None
        self._sheet_id_backup   = None
        self.slack_team_id      = None
        self.slack_channel      = None
        self.verbose            = False
        self.file_id            = file_id
        self.sheet_title        = 'Jira Data'
        self.backup_sheet_title = 'original_jira_data'
        self.headers            = []
        self.gsuite_secret_id   = gsuite_secret_id
        self.elastic_secret_id  = 'elastic-jira-dev-2'
        if not self.gsuite_secret_id:
            self.gsuite_secret_id = 'gsuite_gsbot_user'


    # Helper methods
    # def elastic(self):
    #     if self._elastic is None:
    #         self._elastic =  API_Issues()
    #     return self._elastic

    def jira(self):
        if self._jira is None:
            self._jira = API_Jira()
        return self._jira

    def jira_rest(self):
        if self._jira_rest is None:
            self._jira_rest = API_Jira_Rest()
        return self._jira_rest


    def gsheets(self):
        if self._gsheets is None:
            self._gsheets = GSheets(gsuite_secret_id=self.gsuite_secret_id)
        return self._gsheets

    def sheet_name(self):
        if self._sheet_name is None:
            sheets = self.gsheets().sheets_properties_by_title(self.file_id)
            if sheets:
                if self.sheet_title not in list(set(sheets)):
                    self.gsheets().sheets_add_sheet(self.file_id, self.sheet_title)
                self._sheet_name = self.sheet_title
        return self._sheet_name

    def sheet_name_backup(self):
        if self._sheet_name_backup is None:
            sheets = self.gsheets().sheets_properties_by_title(self.file_id)
            if self.backup_sheet_title not in list(set(sheets)):
                self.gsheets().sheets_add_sheet(self.file_id, self.backup_sheet_title)
            self._sheet_name_backup = self.backup_sheet_title
        return self._sheet_name_backup

    def sheet_id(self):
        if self._sheet_id is None:
            sheets = self.gsheets().sheets_properties_by_title(self.file_id)
            if self.sheet_title in list(set(sheets)):
                self._sheet_id = sheets.get(self.sheet_title).get('sheetId')
        return self._sheet_id

    def sheet_id_backup(self):
        if self._sheet_id_backup is None:
            sheets = self.gsheets().sheets_properties_by_title(self.file_id)
            if self.backup_sheet_title in list(set(sheets)):
                self._sheet_id_backup = sheets.get(self.backup_sheet_title).get('sheetId')
        return self._sheet_id_backup

    def set_slack_support(self, team_id, channel):
        self.slack_team_id = team_id
        self.slack_channel = channel

    def log_message(self, message):
        if self.verbose:
            Dev.pprint(message)
        if self.slack_team_id and self.slack_channel:
            slack_message(message, self.slack_team_id, self.slack_channel)
        return message

    def log_error(self, error):
        return self.log_message(':red_circle: {0}'.format(error))

    # main methods

    def convert_sheet_data_to_raw_data(self,sheet_data):
        raw_data = [self.headers]
        for item in sheet_data:
            row = []
            for header in self.headers:
                row.append(item.get(header))
            raw_data.append(row)
        return raw_data

    def get_sheet_raw_data(self, sheet_name):
        return self.gsheets().get_values(self.file_id, sheet_name)

    #@save_result_to_local_cache
    #@use_local_cache_if_available
    def get_sheet_data(self,sheet_name):
        rows = self.get_sheet_raw_data(sheet_name)
        if rows:
            self.headers = rows.pop(0)
            data = []
            for row_index, row in enumerate(rows):
                item = { 'index':row_index}
                for header_index, header in enumerate(self.headers):
                    if header_index >= len(row):
                        value = None
                    else:
                        value  = row[header_index]
                    if header == 'Jira Link' and len(row) > 0:
                        value = '=HYPERLINK("https://jira.photobox.com/browse/{0}","{0}")'.format(row[0])
                    item[header] = value
                data.append(item)
            return data

    def color_code_cells_based_on_diff_status(self, diff_cells):
        sheet_id = self.sheet_id()

        requests = []
        for diff_cell in diff_cells:
            col    = diff_cell.get('col_index')
            row    = diff_cell.get('row_index')
            status = diff_cell.get('status')
            #if status == 'same'          : requests.append(self.gsheets().request_cell_set_background_color(sheet_id, col ,row, 0.5 ,1.0 ,0.5))
            if status == 'sheet_change'  : requests.append(self.gsheets().request_cell_set_background_color(sheet_id, col ,row, 1.0 ,0.5 ,0.5))
            if status == 'jira_change'   : requests.append(self.gsheets().request_cell_set_background_color(sheet_id, col ,row, 1.0 ,0.5 ,0.5))
            if status == 'other'         : requests.append(self.gsheets().request_cell_set_background_color(sheet_id, col ,row, 0.5 ,0.5 ,0.5))
            if status == 'jira-save-ok'  : requests.append(self.gsheets().request_cell_set_background_color(sheet_id, col ,row, 0.0 ,0.5 ,0.5))
            if status == 'jira-save-fail': requests.append(self.gsheets().request_cell_set_background_color(sheet_id, col ,row, 1.0 ,0.0 ,0.0))

        return self.gsheets().execute_requests(self.file_id, requests)

    def diff_cells(self):
        sheet_data = self.get_sheet_data(self.sheet_name())
        backup_data = self.get_sheet_data(self.sheet_name_backup())
        self.log_message("In API_Jira_Sheets_Sync.diff_cells: got sheet_data and backup_data")
        if backup_data is None:
            self.log_error("Error: Backup data is not setup, please load the data again")
            return None
        jira_issues = self.get_jira_issues_in_sheet_data(sheet_data)
        self.log_message("In API_Jira_Sheets_Sync.diff_cells: got {0} issues from Jira".format(len(jira_issues)))
        diff_data = self.diff_sheet_data_with_jira_data(sheet_data, backup_data, jira_issues)
        self.log_message("In API_Jira_Sheets_Sync.diff_cells: got {0} items in diff_data".format(len(diff_data)))
        return diff_data

    def diff_sheet(self):
        diff_cells = self.diff_cells()
        if diff_cells:
            self.color_code_cells_based_on_diff_status(diff_cells)
            return "diff completed..."
        return "Error, could not calculate diff_cells (try reloading the data)"

    def diff_sheet_data_with_jira_data(self,sheet_data, backup_data, jira_data):
        if backup_data is None or jira_data is None:
            return None
        diff_cells = []
        if self.verbose:
            print()
        for row_index, row in enumerate(sheet_data):
            try:
                for header_index, header in enumerate(self.headers):
                    key = row['Key']
                    jira_issue    = jira_data.get(key)
                    backup_issue  = backup_data[row_index]
                    if jira_issue and backup_issue:
                        if header in ['Jira Link']: continue
                        sheet_value  = row.get(header)
                        backup_value = backup_issue.get(header, '')
                        jira_value   = jira_issue.get(header,'')

                        if sheet_value:
                            diff_cell = {
                                            'key'         : key          ,
                                            'field'       : header       ,
                                            'row_index'   : row_index + 1,
                                            'col_index'   : header_index ,
                                            'sheet_value' : sheet_value  ,
                                            'backup_value': backup_value ,
                                            'jira_value'  : jira_value   ,

                                        }
                            if   sheet_value == jira_value and sheet_value == backup_value: diff_cell['status'] = 'same'
                            elif sheet_value != jira_value and sheet_value == backup_value: diff_cell['status'] = 'jira_change'
                            elif sheet_value != jira_value and jira_value  == ''          : diff_cell['status'] = 'sheet_change'
                            elif sheet_value != jira_value and jira_value  is None        : diff_cell['status'] = 'sheet_change'
                            elif sheet_value != jira_value and jira_value  == backup_value: diff_cell['status'] = 'sheet_change'
                            else: diff_cell['status'] = 'other'
                            #print("{0:10} {1:20} {2:20} {3:20}".format(key, header, sheet_value, jira_value))
                            diff_cells.append(diff_cell)
            except Exception as error:
                Dev.pprint("Error in row {0}: {1}".format(row_index,row))
        return diff_cells

    def update_sheet_data_with_jira_data(self,sheet_data):
        for item in sheet_data:
            issue = self.get_issue_data(item.get('Key'))
            if issue:
                for column_id in set(item):
                    value = issue.get(column_id)
                    if value:
                        item[column_id] = value

    def get_jira_issues_in_sheet_data(self, sheet_data):
        keys = [row.get('Key') for row in sheet_data if row.get('Key') != '']
        #issues = self.jira().issues(keys)
        issues = self.jira_rest().issues(keys)
        return issues

    #@use_local_cache_if_available
    def get_issues_data(self,issues_id):
        return self.jira_rest().issues(issues_id)

    #@use_local_cache_if_available
    def get_issue_data(self,issue_id):
        return self.jira_rest().issue(issue_id)
        #return self.jira().issue(issue_id)

    def update_file_with_raw_data(self,raw_data,sheet_name):
        self.gsheets().set_values(self.file_id, sheet_name, raw_data)

    def load_data_from_jira(self):
        sheet_name = self.sheet_name()
        if not sheet_name:
            return "Error: in `load_data_from_jira` the `sheet_name` value could not be calculated from the current file Id: {0}".format(self.file_id)
        sheet_data = self.get_sheet_data(sheet_name)
        if sheet_data:
            try:
                self.update_sheet_data_with_jira_data(sheet_data)
                raw_data = self.convert_sheet_data_to_raw_data(sheet_data)
                self.update_file_with_raw_data(raw_data, self.sheet_name())             # add data to both
                self.update_file_with_raw_data(raw_data, self.sheet_name_backup())      # first one and the gsbot backup one (to be used to calculate update diffs)
                return "loaded data from jira completed...."
            except Exception as error:
                return "Error in load_data_from_jira: {0}".format(error)
        return "Error: no data for file_id: {0}".format(self.file_id)

    def sync_sheet(self):
        self.log_message("In API_Jira_Sheets_Sync.sync_sheet")
        try:
            diff_cells = self.diff_cells()
            if diff_cells:
                self.log_message("In API_Jira_Sheets_Sync.sync_sheet: about to sync_data_between_jira_and_sheet")
                self.sync_data_between_jira_and_sheet(diff_cells)
                self.log_message("In API_Jira_Sheets_Sync.sync_sheet: completed sync_data_between_jira_and_sheet")
                return "sync data with Jira completed...."
            return self.log_error("Error, could not calculate diff_cells (try reloading the data)")
        except Exception as error:
            return self.log_error("Error in sync_sheet: {0}".format(error))

    def sync_data_between_jira_and_sheet(self,diff_cells):
        #fields_to_update = {}
        for item in diff_cells:
            status = item.get('status')
            if status == 'sheet_change':
                key   = item.get('key')
                field = item.get('field')
                value = item.get('sheet_value')
                #fields_to_update[key][field] = value
                message ='[Jira update] for jira issue `{0}` updating field `{1}` with value `{2}` '.format(key, field, value)
                if self.verbose:
                    print(message)
                result = self.jira_rest().issue_update_field(key, field, value)
                self.update_backup_data_with_new_jira_value(item)
                if result:
                    item['status'] = 'jira-save-ok'
                else:
                    item['status'] = 'jira-save-fail'
            if status == 'jira_change':
                key = item.get('key')
                field = item.get('field')
                value = item.get('sheet_value')
                if self.verbose:
                    Dev.pprint('need to update sheet, for jira issue `{0}` field `{1}` value `{2}` '.format(key, field, value))
        self.color_code_cells_based_on_diff_status(diff_cells)


    def update_backup_data_with_new_jira_value(self,item):
        sheet_id = self.sheet_id_backup()
        requests = [self.gsheets().request_cell_set_value(sheet_id,item.get('col_index'), item.get('row_index'),item.get('sheet_value'))]

        return self.gsheets().execute_requests(self.file_id,requests)

    #@use_local_cache_if_available
    #@save_result_to_local_cache
    def get_graph_nodes(self,graph_name):
        from osbot_jira.api.graph.Lambda_Graph import Lambda_Graph
        graph = Lambda_Graph().get_gs_graph___by_name(graph_name)
        if graph:
            return sorted(graph.nodes)
        return []

    def create_sheet_from_graph(self, graph_name,domain, folder):
        title = "Data for graph - {0}".format(graph_name)
        self.file_id = self.gsheets().create_and_share_with_domain(title, domain,folder)
        headers = ['Key', 'Jira Link', 'Summary', 'Status', 'Risk Rating', 'Issue Type']

        self.create_sheet_from_graph_with_headers(graph_name, headers)
        self.gsheets().sheets_delete_sheet(self.file_id, 0)
        return self.gsheets().gdrive.file_weblink(self.file_id)

    def create_sheet_from_graph_with_headers(self,graph_name, headers=None):
        issues_ids = self.get_graph_nodes(graph_name)
        issues     = self.get_issues_data(issues_ids)

        if headers is None:
            headers    = ['Key']
            for key,values in issues.items():
                for header in values:
                    headers.append(header)
            headers = list(set(headers))
        values = [headers]
        for key,issue in issues.items():
            row = []
            for index in range(0,len(headers)):
                header = headers[index]
                if header == 'Jira Link':
                    value = '=HYPERLINK("https://jira.photobox.com/browse/{0}","{0}")'.format(key)
                else:
                    value = issue.get(header,'')
                row.append(value)

            values.append(row)

        sheet_name = self.sheet_name()
        self.gsheets().set_values(self.file_id,sheet_name,values)
        return self.gsheets().gdrive.file_weblink(self.file_id)



from osbot_gsuite.apis.GDrive_Activity import GDrive_Activity
from osbot_gsuite.apis.GSheet import GSheet
from osbot_gsuite.apis.GSheets import GSheets
from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Json import json_load_file

TEMP_ACTIVITIES_PATH = '/tmp/gdrive_activity.json'


class Incident_Mass_Delete:

    def __init__(self):
        self.gdrive_activity = GDrive_Activity()
        return

    def raw_activities(self):
        return json_load_file(TEMP_ACTIVITIES_PATH)

    def actions(self, max=2):
        raw_activities = self.raw_activities()
        if max >0:
            raw_activities = raw_activities[:max]
        actions = self.gdrive_activity.parse_activities(raw_activities)
        self.resolve_actions_persons(actions)
        return actions

    def resolve_actions_persons(self, actions):
        persons = self.gdrive_activity.persons_details_from_actions(actions)
        for action in actions:
            person_id              = action.get('person_id')
            person                 = persons.get(person_id)
            action['display_name'] = person.get('display_name', 'unknown')

    def gsheet_data_from_actions(self, actions):
        rows = []
        for action in actions:
            target      = action.get('target'      )
            action_name = action.get('action_name' )
            action_data = action.get('action_data')
            target_data = action.get('target_data')
            row = {"Date"         : action.get('date'        ),
                   "Time"         : action.get('time'        ),
                   "Name"         : action.get('display_name'),
                   "Action"       : action_name               ,
                   #"action_data": action.get('action_data'),
                   "Target"       : target                    ,
                   'Target Title' : target_data.get('title'  ),
                   #"target_data": action.get('target_data'),
                   "Person_id"    : action.get('person_id'),
                   }

            rows.append(row)
        return rows

    def gsheet_send_actions_data(self, actions, file_id, sheet_name='Sheet1'):
        gsheets = GSheets()
        gsheet_data = self.gsheet_data_from_actions(actions)
        return gsheets.set_sheet_data(file_id, sheet_name, gsheet_data)



from osbot_aws.helpers.Lambda_Helpers import log_to_elk
from osbot_aws.Dependencies import load_dependencies


def run(event, context):
    try:
        load_dependencies(["gmail", "elastic-slack", 'requests'])                                    # load dependency (download and unzip if first run)
        from osbot_gsuite.apis.handlers.Lambda_Slides import Lambda_Slides
        return Lambda_Slides().handle_lambda_event(event)                                   # invoke lambda handler from Lambda_Slides class

        #return "200 OK"

    except Exception as error:
        message = "[lambda_slides] Error: {0}".format(error)
        log_to_elk(message, level='error')
        return message


from osbot_gsuite.apis.GDrive import GDrive
from osbot_gsuite.apis.GSuite import GSuite


class GDocs:

    def __init__(self, gsuite_secret_id=None):
        self.docs   = GSuite(gsuite_secret_id).docs_v1().documents()
        self.gdrive = GDrive(gsuite_secret_id)


    def execute_requests(self, file_id, requests):
        body = {'requests': requests}
        return self.docs.batchUpdate(documentId= file_id, body=body).execute()


from googleapiclient.errors import HttpError

from osbot_utils.utils.Dev import pprint


class GDoc_Named_Range:

    def __init__(self, gdoc, named_range_name):
        self.gdoc             = gdoc
        self.named_range_name = named_range_name
        self.replies          = []
        self.start_index      = None
        self.end_index        = None

    def __enter__(self):
        self.set_range_indexes()
        self.gdoc.requests = []
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        print(f"[GDoc_Named_Range] executing {len(self.gdoc.requests)} requests")
        if exc_type:
            return False
        try:
            self.replies = self.gdoc.commit()
        except HttpError as error:
            print('\n[GDoc commit() Error]:', error.error_details)
            print(type(error))
            raise Exception(error)
        return True

    def add_text(self, text, location):
        if self.start_index < location < self.end_index:
            self.gdoc.insert_text(text=text, location=location)
        return self

    def append_image(self, image_url,width=None, height=None):
        if self.end_index:
            self.gdoc.add_request_insert_inline_image_from_image_url(image_url=image_url, location=self.end_index, width=width, height=height)
            self.end_index += 1
            self.reset_named_range()            # need to reset because the end index needs to be adjusted
        return self

    def append_table(self, rows, columns):
        if self.end_index:
            self.gdoc.add_request_insert_table(rows=rows, columns=columns, location=self.end_index)

    def append_text(self, text):
        if self.end_index and len(text) > 0:
            self.gdoc.insert_text(text=text, location=self.end_index)
            self.end_index += len(text)
            self.reset_named_range()            # need to reset because the end index needs to be adjusted
        return self

    def append_text_with_bold(self, text, value=True):
        text_style = {"bold": value}
        return self.append_text_with_style(text, text_style=text_style)

    def append_text_with_named_style(self, text, named_style):
        paragraph_style = {"named_style": named_style}
        return self.append_text_with_style(text, paragraph_style=paragraph_style)

    def append_text_with_style(self, text, text_style=None, paragraph_style=None):
        start_index = self.end_index
        self.append_text(text)
        end_index   = self.end_index
        range = { "start_index": start_index ,
                  "end_index"  : end_index   }
        if text_style:
            self.set_text_style_to_range(range, text_style)
        if paragraph_style:
            self.set_paragraph_style_to_range(range, paragraph_style)
        return self



    def background_color(self, color):
        text_style = {"background_color": color}
        #self.gdoc.add_request_text_style_to_range(self.range(), text_style)
        return self.set_text_style_to_range(self.range(), text_style)

    def clear(self):
        return self.set_text('\n')      # if we set it to an empty string, the named range will be deleted

    def info(self):
        return { "name"        : self.named_range_name,
                 "start_index" : self.start_index     ,
                 "end_index"   : self.end_index       }

    def insert_text(self, text):
        self.gdoc.insert_text(text=text, location=self.start_index)
        self.end_index += len(text)
        self.reset_named_range()            # need to reset because the start index needs to be adjusted
        return self

    def page_break(self):
        self.gdoc.add_request_page_break(self.end_index)
        return self

    def new_line(self):
        self.append_text('\n')
        return self

    def range(self):
        return {"start_index": self.start_index, "end_index": self.end_index}

    # bug: named_range is not correctly fixed when there are more than one match
    # todo: see if there is a way to detect the number of matches
    def replace(self, text_to_find, text_to_replace):
        self.gdoc.add_request_replace_text(text_to_find=text_to_find, text_to_replace=text_to_replace)
        self.end_index = self.end_index + len(text_to_replace) - len(text_to_find)
        #print(self.end_index)
        self.reset_named_range()
        return self

    def reset_named_range(self):
        self.gdoc.add_request_named_range_reset(name=self.named_range_name,start_index=self.start_index, end_index=self.end_index)
        return self

    def set_paragraph_style_to_range(self, range, paragraph_style):
        return self.gdoc.add_request_paragraph_style_to_range(range, paragraph_style)

    def set_range_indexes(self):
        ranges           = self.gdoc.named_range_ranges(self.named_range_name)
        data             = self.gdoc.ranges_start_and_end(ranges)
        self.start_index = data.get('start_index')
        self.end_index   = data.get('end_index')
        return self

    def set_text(self, text):
        if self.start_index:
            self.gdoc.add_request_replace_named_range_content(name=self.named_range_name, text=text)
            self.end_index = self.start_index + len(text)
        return self

    def set_text_style_to_range(self, range, text_style):
        return self.gdoc.add_request_text_style_to_range(range, text_style)

    def __repr__(self):
        info = self.info()
        return f"{info.get('name')} - from {info.get('start_index')} to {info.get('end_index')}"



from dateutil import parser

from osbot_gsuite.apis.GDrive import GDrive
from osbot_gsuite.apis.GPeople import GPeople
from osbot_gsuite.apis.GSuite import GSuite
from osbot_utils.decorators.methods.cache_on_self import cache_on_self
from osbot_utils.decorators.methods.cache_on_tmp import cache_on_tmp
from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Json import json_save_file
from osbot_utils.utils.Misc import list_set
from osbot_utils.utils.Python_Logger import logger_info

TEMP_ACTIVITIES_PATH = '/tmp/gdrive_activity.json'

class GDrive_Activity:

    def __init__(self):
        self.log_info = logger_info()
        self.gpeople = GPeople()

    @cache_on_self
    def activity(self):
        return GSuite().drive_activity_v2().activity()

    def gdrive(self):
        return GDrive()

    #@cache_on_tmp()
    def query(self, item=None, page_size=2, recursive=False, loop=1):
        body           = { 'pageSize': page_size ,
                           'pageToken': None}
        request_params = { 'body'    : body      }
        if item:
            if recursive:
                body['ancestorName'] = f'items/{item}'
            else:
                body['itemName'] = f'items/{item}'

        all_activities = []
        for i in range(loop):
            self.log_info(f"making request #{i+1} of {loop}")
            result = self.activity().query(**request_params).execute()
            activities      = result.get('activities', [])
            next_page_token = result.get('nextPageToken')
            self.log_info(f"got {len(activities)} activities")
            all_activities.extend(activities)
            if next_page_token is None:
                break
            body['pageToken'] = next_page_token
        self.save_all_activities(all_activities)
        #return activities
        return self.parse_activities(all_activities)

    def parse_activity(self, activity):
        actions               = activity.get('actions')
        actors                = activity.get('actors')
        primary_action_detail = activity.get('primaryActionDetail')
        targets               = activity.get('targets')
        time_stamp            = activity.get('timestamp')


        # assert expectations of data analysis below
        assert list_set(activity) == ['actions', 'actors', 'primaryActionDetail', 'targets', 'timestamp']
        assert len(actors) == 1

        actor       = actors.pop()
        user        = actor.get('user')
        known_user  = user.get('knownUser')
        person_name = known_user.get('personName')

        assert list_set(actor     ) == ['user'      ]
        assert list_set(user      ) == ['knownUser' ]
        assert list_set(known_user) == ['personName']
        assert len(person_name.split('/')) == 2
        assert len(targets)                == 1

        targets    = targets.pop()                              # since target has name value pairs inside it, and we have already asserted that it is only one value (via the assert above)
        person_id  = person_name.split('/').pop()
        date_time  = parser.parse(timestr=time_stamp)
        date       = str(date_time.date())
        time       = f'{date_time.hour}:{date_time.minute}.{date_time.second}'
        activities = []

        for action in actions:
            assert list_set(action) == ['detail']
            for action_name,action_data in action.get('detail').items():
                for target, target_data in targets.items():
                    activity = {'date'        : date        ,
                                'time'        : time        ,
                                'person_id'   : person_id   ,
                                'action_name' : action_name ,
                                'action_data' : action_data ,
                                'target'      : target      ,
                                'target_data' : target_data }
            activities.append(activity)
        return activities

    def parse_activities(self, activities):
        self.log_info(f'parsing {len(activities)} activities')
        actions = []
        for activity in activities:
            actions.extend(self.parse_activity(activity))
        self.log_info(f'which were expanded into {len(actions)} actions')
        return actions
        #return [self.parse_activity(activity for activity in activities)]

    def persons_details_from_actions(self, actions):
        persons = {}
        for action in actions:
            person_id = action.get('person_id')
            if person_id not in persons:
                persons[person_id] = {}

        self.log_info(f"There are {len(persons)} persons to resolve")
        for person_id in persons:
            try:
                persons[person_id] = self.gpeople.person(person_id)
            except Exception as error:
                self.log_info(f"Error resolving person id: {person_id} - {error}")
                persons[person_id] = { }
        return persons

    def save_all_activities(self, activities):
        result = json_save_file(activities, TEMP_ACTIVITIES_PATH, pretty=False)
        self.log_info(f'saved {len(activities)} activities to {TEMP_ACTIVITIES_PATH}')


import random

from osbot_utils.decorators.methods.cache_on_self import cache_on_self
from osbot_utils.testing.Duration import Duration


class Alignment:
    ALIGNMENT_UNSPECIFIED   = "ALIGNMENT_UNSPECIFIED"
    START                   = "START"
    CENTER                  = "CENTER"
    END                     = "END"
    JUSTIFIED               = "JUSTIFIED"


class Baseline_Offset:
    UNSPECIFIED = 'BASELINE_OFFSET_UNSPECIFIED'
    NONE        = 'NONE'
    SUPERSCRIPT = 'SUPERSCRIPT'
    SUBSCRIPT   = 'SUBSCRIPT'

class Direction:
    UNSPECIFIED   = "CONTENT_DIRECTION_UNSPECIFIED"
    LEFT_TO_RIGHT = "LEFT_TO_RIGHT"
    RIGHT_TO_LEFT = "RIGHT_TO_LEFT"

class RGB:                                                                  # see https://kierandixon.com/google-sheets-colors/ for list of colors
    AZURE       = { "red": 0      , "green": 0.5    , "blue": 1      }
    BLACK       = { "red": 0      , "green": 0      , "blue": 0      }
    BLUE        = { "red": 0      , "green": 0      , "blue": 1      }
    BLUE_L2     = { "red": 159/255, "green": 197/255, "blue": 232/255} # light blue 2  | 9fc5e8 | 159, 197, 232
    BLUE_L3     = { "red": 207/255, "green": 226/255, "blue": 243/255} # light blue 3  | cfe2f3 | 207, 226, 243
    CYAN        = { "red":   0/255, "green": 255/255, "blue": 255/255} # cyan          | 00ffff |   0, 255, 255
    CYAN_L1     = { "red": 118/255, "green": 165/255, "blue": 175/255} # light cyan  1 | 76a5af | 118, 165, 175
    CYAN_L2     = { "red": 162/255, "green": 196/255, "blue": 201/255} # light cyan  2 | a2c4c9 | 162, 196, 201
    CYAN_L3     = { "red": 208/255, "green": 224/255, "blue": 227/255} # light cyan  3 | d0e0e3 | 208, 224, 227
    #DARK_BLUE_2 = { "red": 0.04   , "green": 0.33   , "blue": 0.58   }
    GREEN       = { "red": 0      , "green": 1      , "blue": 0      }
    GREEN_L1    = { "red": 147/255, "green": 196/255, "blue": 125/255} # light green 1 | 93c47d | 147, 196, 125
    GREEN_L2    = { "red": 182/255, "green": 215/255, "blue": 168/255} # light green 2 | b6d7a8 | 182, 215, 168
    GREEN_L3    = { "red": 217/255, "green": 234/255, "blue": 211/255} # light green 3 | d9ead3 | 217, 234, 211
    GREY        = { "red": 204/255, "green": 204/255, "blue": 204/255} # grey          | ccccc  | 204, 204, 204
    GREY_L1     = { "red": 217/255, "green": 217/255, "blue": 217/255} # light grey  1 | d9d9d9 | 217, 217, 217
    GREY_L2     = { "red": 239/255, "green": 239/255, "blue": 239/255} # light grey  2 | efefef | 239, 239, 239
    GREY_L3     = { "red": 243/255, "green": 243/255, "blue": 243/255} # light grey  3 | f3f3f3 | 243, 243, 243
    GREY_D1     = { "red": 183/255, "green": 183/255, "blue": 183/255} # Dark grey  1  | b7b7b7 | 183, 183, 183
    GREY_D2     = { "red": 153/255, "green": 153/255, "blue": 153/255} # Dark grey  2  | 999999 | 153, 153, 153
    GREY_D3     = { "red": 102/255, "green": 102/255, "blue": 102/255} # Dark grey  3  | 666666 | 102, 102, 102
    GREY_D4     = { "red":  67/255, "green":  67/255, "blue":  67/255} # Dark grey  4  | 434343 |  67,  67,  67
    RED         = { "red": 1      , "green": 0      , "blue": 0      }
    RED_L3      = { "red": 244/255, "green": 204/255, "blue": 204/255} # light red 3   | f4cccc | 244, 204, 204
    WHITE       = { "red": 1      , "green": 1      , "blue": 1      }

    @cache_on_self
    def all_color_keys(self):
        return list(key for key, value in RGB.__dict__.items() if isinstance(value, dict))

    @cache_on_self
    def all_color_values(self):
        return list(value for key, value in RGB.__dict__.items() if isinstance(value, dict))

    def random_color_value(self):
        return random.choice(self.all_color_values())


        #return set(RGB)

class Border_Style:
    DOTTED	        = "DOTTED"
    DASHED          = "DASHED"
    SOLID           = "SOLID"
    SOLID_MEDIUM    = "SOLID_MEDIUM"
    SOLID_THICK     = "SOLID_THICK"
    NONE            = "NONE"
    DOUBLE          = "DOUBLE"

class Content_Alignment:
    UNSPECIFIED = "CONTENT_ALIGNMENT_UNSPECIFIED"
    TOP         = "TOP"
    MIDDLE      = "MIDDLE"
    BOTTOM      = "BOTTOM"

class Dash_Style:
    UNSPECIFIED = "DASH_STYLE_UNSPECIFIED"
    SOLID       = "SOLID"
    DOT         = "DOT"
    DASH        = "DASH"

class Font_Family:
    ARIAL           = "Arial"
    CAVEAT          = "Caveat"
    OSWALD          = "Oswald"
    VERDANA         = "Verdana"

class Merge_Type:
    MERGE_ALL     = "MERGE_ALL"
    MERGE_COLUMNS = "MERGE_COLUMNS"
    MERGE_ROWS    = "MERGE_ROWS"

class Named_Style:
    UNSPECIFIED = "NAMED_STYLE_TYPE_UNSPECIFIED"
    NORMAL_TEXT = "NORMAL_TEXT"
    TITLE	    = "TITLE"
    SUBTITLE    = "SUBTITLE"
    HEADING_1   = "HEADING_1"
    HEADING_2   = "HEADING_2"
    HEADING_3   = "HEADING_3"
    HEADING_4   = "HEADING_4"
    HEADING_5   = "HEADING_5"
    HEADING_6   = "HEADING_6"

class Spacing_Mode:
    SPACING_MODE_UNSPECIFIED = "SPACING_MODE_UNSPECIFIED"
    NEVER_COLLAPSE           = "NEVER_COLLAPSE"
    COLLAPSE_LISTS           = "COLLAPSE_LISTS"

class Width_Type:
    UNSPECIFIED            = "WIDTH_TYPE_UNSPECIFIED"
    EVENLY_DISTRIBUTED     = "EVENLY_DISTRIBUTED"
    FIXED_WIDTH            = "FIXED_WIDTH"

class Wrap_Strategy:
    OVERFLOW_CELL = "OVERFLOW_CELL"
    CLIP          = "CLIP"
    WRAP          = "WRAP"



from osbot_gsuite.apis.GCalendar import GCalendar

class Calendar_Commands:
    gsuite_secret_id = 'gsuite_gsbot_user'

    @staticmethod
    def gs_team(team_id=None, channel=None, params=None):
        calendar = GCalendar(Calendar_Commands.gsuite_secret_id)
        return calendar.gs_team()


from osbot_gsuite.apis.GDrive import GDrive
from osbot_utils.decorators.lists.index_by import index_by
from osbot_utils.utils.Files import Files


class GDrive_Folder:

    def __init__(self, folder_name=None, folder_id=None):
        self.folder_name = folder_name
        self.folder_id   = folder_id
        self.gdrive = GDrive()

    def resolve_folder_id(self):
        if self.folder_id is None:
            mime_type = 'application/vnd.google-apps.folder'
            result = self.gdrive.find_by_name(self.folder_name, mime_type)
            if result:
                self.folder_id = result.get('id')
        return self

    def file_info(self, file_id):
        return self.gdrive.file_metadata(file_id=file_id, fields='*')

    def file_delete(self,file_id):
        return self.gdrive.file_delete(file_id)

    def file_upload(self, path, mime_type):
        print(f'uploading file: {path}')
        self.resolve_folder_id()
        if self.folder_id:
            return self.gdrive.file_upload(local_file=path, mime_type=mime_type, folder_id=self.folder_id)

    def file_upload_png(self, path):
        png_mime_type = 'image/png'
        return self.file_upload(path=path, mime_type=png_mime_type)

    @index_by
    def files(self, size=100, fields="files(id,name)"):
        return self.gdrive.files_in_folder(folder_id=self.folder_id, size=size, fields=fields)

        #file_id = self.graph_id_in_gdrive(file_name)

        #if file_id:  # if the file exists, delete it before uploading the new version
        #    Dev.pprint('deleting file: {0}'.format(file_id))
        #    self.gdrive.file_delete(file_id)

    def folder_create(self):
        self.resolve_folder_id()
        if self.folder_name and self.folder_id is None:
            return self.gdrive.folder_create(self.folder_name)


from osbot_aws.helpers.Lambda_Helpers import log_to_elk
from osbot_aws.Dependencies import load_dependency


def run(event, context):
    try:
        load_dependency("gmail")
        #load_dependency("elastic-slack")                                      # load dependency (download and unzip if first run)
        from osbot_gsuite.apis.handlers.Lambda_Calendar import Lambda_Calendar
        return Lambda_Calendar().handle_lambda_event(event)                    # invoke lambda handler from Lambda_Slides class

    except Exception as error:
        message = "[lambda_gcalendar] Error: {0}".format(error)
        log_to_elk(message, level='error')
        return message


from googleapiclient.discovery import Resource
from googleapiclient.http    import MediaFileUpload
from osbot_utils.base_classes.Type_Safe import Type_Safe

from osbot_gsuite.gsuite.GSuite                     import GSuite
from osbot_utils.decorators.methods.cache_on_self   import cache_on_self
from osbot_utils.utils.Dev                          import Dev
from osbot_utils.utils.Files                        import Files


class GDrive(Type_Safe):
    gsuite: GSuite

    @cache_on_self
    def drive_v3(self) -> Resource:                                       # todo refactor into drive_v3
        return self.gsuite.drive_v3()

    @cache_on_self
    def files(self):
        return self.drive_v3().files()

    @cache_on_self
    def permissions(self):
        return self.drive_v3().permissions()

    def execute(self, command):
        try:
            return command.execute()
        except Exception as error:
            Dev.pprint(error)                   # add better error handling log capture
            return None

    def folder_create(self, folder_name):
        return self.file_create(file_type='application/vnd.google-apps.folder', title=folder_name)

    def file_create(self, file_type, title, folder=None):
        file_metadata = {
            "mimeType": file_type,
            "name": title
        }
        if folder:
            file_metadata["parents"] =  [folder]
        file = self.files().create(body=file_metadata, fields='id').execute()
        return file.get('id')

    def file_export(self, file_Id):
        return self.files().export(fileId=file_Id, mimeType='application/pdf').execute()

    def file_export_as_pdf_to(self,file_id,target_file):
        pdf_data = self.file_export(file_id)
        with open(target_file, "wb") as fh:
            fh.write(pdf_data)
        return target_file

    def file_metadata(self, file_id, fields = "id,name"):
        return self.execute(self.files().get(fileId = file_id, fields=fields))

    def file_metadata_update(self, file_id, changes):
        return self.files().update(fileId=file_id, body=changes).execute()

    def file_delete(self, file_id):
        if file_id:
            self.files().delete(fileId= file_id).execute()

    def file_share_with_domain(self,file_id,domain):
        domain_permission = {
            'type': 'domain',
            'role': 'writer',
            'domain': domain,
        }
        return self.permissions().create(fileId=file_id, body=domain_permission, fields='id').execute()

    def file_update(self, file_id, body):
        return self.files().update(fileId=file_id, body=body).execute()
        # if Files.exists(local_file):
        #     file_metadata = {'name': Files.file_name(local_file)}
        #     media = MediaFileUpload(local_file, mimetype=mime_type)
        #     file = self.files().update(body=file_metadata, media_body=media, fileId= file_id, fields='id').execute()
        #     return file.get('id')
        #return None

    def file_upload(self, local_file, mime_type, folder_id=None):
        if Files.exists(local_file):
            file_metadata = {'name': Files.file_name(local_file), 'parents': [folder_id]}
            media = MediaFileUpload(local_file, mimetype=mime_type)
            file = self.files().create(body=file_metadata, media_body=media, fields='id').execute()
            return file.get('id')
        return None

    def file_weblink(self, file_id):
        return 'https://drive.google.com/open?id={0}'.format(file_id)

    def files_all(self, size, fields="files(id,name)"):
        results = self.files().list(pageSize=size, fields=fields).execute()
        return results.get('files', [])

    def files_in_folder(self, folder_id, size=100, fields="files(id,name)"):
        results = self.files().list(q="parents='{0}' and trashed = false".format(folder_id),pageSize=size, fields=fields).execute()
        return results.get('files', [])

    def find_by_name(self, name, mime_type = None):
        if mime_type:
            query = "name = '{0}' and mimeType = '{1}'".format(name,mime_type)
        else:
            query = "name = '{0}'".format(name)
        results = self.execute(self.files().list(q=query))  # , fields="files(id,name)"))
        if results and len(results.get('files')) > 0:
            return results.get('files').pop()
        return None

    def find_by_mime_type(self, mime_type):
        results = self.execute(self.files().list(q="mimeType = '{0}'".format(mime_type), fields="files(id,name)"))
        if results:
            return results.get('files', [])

    def set_file_title(self, file_id, new_title):
        return self.file_metadata_update(file_id, {"name" : new_title })



import os

from google_auth_oauthlib.flow      import InstalledAppFlow
from googleapiclient.discovery import build
from oauth2client                   import file, client
from oauth2client.tools             import argparser, run_flow
from google.oauth2.credentials      import Credentials
from google.auth.transport.requests import Request

from osbot_aws.apis.Secrets  import Secrets
from osbot_gsuite.apis.GSheets import GSheets
from osbot_gsuite.apis.GSlides import GSlides
from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Files import file_not_exists, file_contents, file_create, path_combine
from osbot_utils.utils.Json import json_file_contents, file_create_json, json_dumps


class GSuite_Setup():

    def __init__(self):
        self.secret_id_gsuite_client_secret = 'gwbot_gsuite_client_secret'
        self.secret_id_gsuite_token         = 'gwbot_gsuite_token_2'

        #self.secret_id_with_credentials     = 'gsuite_token'
        #self.secret_id_to_create            = 'gsuite_gsbot_user'
        self.file_credentials               = '/tmp/gsuite_client_secret_{0}.json'.format(self.secret_id_gsuite_client_secret)  # todo: refactor to make it OS independent
        self.file_token                     = '/tmp/gmail_credential_{0}.json'    .format(self.secret_id_gsuite_token)

    # Create credentials at https://console.developers.google.com/apis/api/docs.googleapis.com/overview
    # and save them to local disk
    def save_gsuite_client_secret_in_aws(self, file_with_credentials):
        if file_not_exists(file_with_credentials):
            raise Exception(f'provided file does not exist: {file_with_credentials}')
        gwbot_gsuite_secret = file_contents(file_with_credentials)

        #if Secrets(self.secret_id_gsuite_client_secret).update_to_json_string(gwbot_gsuite_secret):        # bug: current privs for the aws account used doesn't have update privs for secrets
        #    return True
        if Secrets(self.secret_id_gsuite_client_secret).create(gwbot_gsuite_secret):
           return True

        raise Exception(f'cloud not update secret: {self.secret_id_gsuite_client_secret}')


    # note: double check thta the 'refresh_token' is set on the AWS Secret (if not, it will fail after a couple hours)
    # todo: figure out how long until then main token expires and Auth workflow is required again
    def create_auth_token_using_web_browser_flow(self, scopes):

        client_secret = Secrets(self.secret_id_gsuite_client_secret).value()                     # get client_secret from AWS Secrets
        store_token   = file.Storage(self.file_token)                                            # create storage object to save token

        file_create(self.file_credentials, client_secret)                                        # save it to temp file

        flow  = client.flow_from_clientsecrets(self.file_credentials, scopes)                    # create a gsuite flow object
        flags = argparser.parse_args('--auth_host_name localhost --logging_level INFO'.split())  # configure the use of a localhost server to received the oauth response
        run_flow(flow, store_token, flags)                                                       # open browser and prompt user to follow the OAuth flow, which will save the token data in self.file_token

        token_json = file_contents(self.file_token)                                              # get gsuite token contents
        Secrets(self.secret_id_gsuite_token).create(token_json)                                  # save it on AWS Secret
        # todo: fix code above to use .update instead of .create (current AWS account only has


    def create_auth_token_using_web_browser_flow__store_in_env(self, scopes):

        #client_secret = Secrets(self.secret_id_gsuite_client_secret).value()                     # get client_secret from AWS Secrets
        folder_auth      = '/Users/diniscruz/_dev/_misc_data/'
        gsuite_auth_file = path_combine(folder_auth, 'client_secret____.apps.googleusercontent.com.json')
        file_credentials = path_combine(folder_auth, 'file_credentials.json')
        file_token       = path_combine(folder_auth, 'file_token.json')

        creds = None

        if os.path.exists(file_token):
            creds = Credentials.from_authorized_user_file(file_token, scopes)
        # If there are no (valid) credentials available, let the user log in.
        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())
            else:
                flow = InstalledAppFlow.from_client_secrets_file(gsuite_auth_file, scopes)
                creds = flow.run_local_server(port=55042)
            # Save the credentials for the next run
            with open(file_token, 'w') as token:
                token.write(creds.to_json())

        # service = build('slides', 'v1', credentials=creds)
        # PRESENTATION_ID = "___"
        # # Call the Slides API
        # print('here')
        # gsheets = GSheets()
        #
        # body = {'properties': {'title': "AAAAAA sheet created from API" }}
        # result = gsheets.spreadsheets.create(body=body).execute()
        # pprint(result)
        # return
        # gslides = GSlides()
        # #pprint(gslides.all_presentations())
        # pprint(gslides.presentation_create("test from HBSec Bot"))
        # presentation = service.presentations().get(
        #     presentationId=PRESENTATION_ID).execute()
        # slides = presentation.get('slides')
        #
        # print('The presentation contains {} slides:'.format(len(slides)))
        # for i, slide in enumerate(slides):
        #     print('- Slide #{} contains {} elements.'.format(
        #         i + 1, len(slide.get('pageElements'))))




import datetime

from osbot_gsuite.apis.GSuite import GSuite


class GCalendar:

    def __init__(self, gsuite_secret_id=None):
        self.events = GSuite(gsuite_secret_id).calendar_v3().events()

    def _get_this_weeks_calendar_data(self, calendar_id):

        #now = datetime.datetime.utcnow().isoformat() + 'Z'
        time_min = '2019-03-17T00:00:00Z'
        time_max = '2019-03-23T00:00:00Z'

        events_result = self.events.list(calendarId = calendar_id,
                                         timeMin    = time_min,
                                         timeMax    = time_max,
                                         singleEvents=True,
                                         orderBy='startTime').execute()
        return events_result.get('items', [])

    def gs_team(self):
        calendar_id = 'photobox.com_kkecukq11iksaamp12p5mqdku0@group.calendar.google.com'
        return self._get_this_weeks_calendar_data(calendar_id)

    def gs_cs_team(self):
        calendar_id = 'photobox.com_b45nntbjhc2255c7dvf1l6ragc@group.calendar.google.com'
        return self._get_this_weeks_calendar_data(calendar_id)


from osbot_utils.utils import Misc

from osbot_gsuite.apis.sheets.API_Jira_Sheets_Sync import API_Jira_Sheets_Sync


class API_Jira_Sheets_Create(API_Jira_Sheets_Sync):
    def __init__(self, file_id):
        super().__init__(file_id)

    def sheet_data(self):
        sheet_name = self.sheet_name()
        return self.get_sheet_data(sheet_name)

    def calculate_jira_actions(self,sheet_data):
        jira_actions = []
        for row in sheet_data:
            index = row.get('index')
            if   row.get('Key'                       ) != 'new' : status = 'issue_exists'
            elif Misc.none_or_empty(row, 'Summary'   )          : status = 'skip_issue - Summary value was empty'
            elif Misc.none_or_empty(row, 'Project'   )          : status = 'skip_issue - Project value was empty'
            elif Misc.none_or_empty(row, 'Issue Type')          : status = 'skip_issue - Issue Type value was empty'
            else:
                status = 'create_issue'
            jira_actions.append({'index':index , 'status': status, 'row': row})

            link_type = row.get('Link Type')
            link_id   = row.get('Link Id')
            key       = row.get('Key')
            if row.get('Link Type') and row.get('Link Id'):
                jira_actions.append({'index': index, 'status': 'create_link', 'row': row, 'link_type': link_type , 'link_id': link_id, 'key': key})

        return jira_actions

    def execute_jira_actions(self, jira_actions):
        for jira_action in jira_actions:
            status = jira_action.get('status')
            if status == 'create_issue':
                row        = jira_action.get('row')
                project    = row.get('Project')
                issue_type = row.get('Issue Type')
                summary    = row.get('Summary')
                description = 'Issue created by GSBot'

                try:
                    result = self.jira().issue_create(project,summary, description,issue_type)
                    jira_action['status'  ] = 'issue_created'
                    jira_action['Jira_Key'] =  result.key
                except Exception as error:
                    jira_action['status'] = "Error: {0}".format(error)

            if status == 'create_link':
                link_type = jira_action.get('link_type')
                to_id     = jira_action.get('link_id')
                from_key  = jira_action.get('key')
                try:
                    result = self.jira().issue_add_link(from_key,link_type,to_id)
                    jira_action['status'  ] = 'link_created'
                except Exception as error:
                    jira_action['status'] = "Error: {0}".format(error)

                print(jira_action)

    def update_sheet_with_status(self,jira_actions):
        requests      = []
        sheet_id      = self.sheet_id()
        status_col    = Misc.array_find(self.headers, 'Sync Status')
        key_col       = Misc.array_find(self.headers, 'Key'   )
        jira_link_col = Misc.array_find(self.headers, 'Jira Link')
        if status_col:
            for jira_action in jira_actions:
                row_index = jira_action.get('index') + 1
                status    = jira_action.get('status')

                if status   == 'create_issue' : requests.append(self.gsheets().request_cell_set_background_color(sheet_id, status_col, row_index, 1.0, 1.0, 0.5))
                elif status == 'issue_created': requests.append(self.gsheets().request_cell_set_background_color(sheet_id, status_col, row_index, 0.5, 1.0, 0.5))
                elif status == 'issue_exists' : requests.append(self.gsheets().request_cell_set_background_color(sheet_id, status_col, row_index, 1.0, 1.0, 1.0))
                elif status == 'create_link'  : requests.append(self.gsheets().request_cell_set_value(sheet_id, status_col + 1, row_index, status))
                elif status == 'link_created' : requests.append(self.gsheets().request_cell_set_background_color(sheet_id, status_col +1, row_index, 0.5, 1.0, 0.5))
                else                          : requests.append(self.gsheets().request_cell_set_background_color(sheet_id, status_col, row_index, 1.0, 0.5, 0.5))

                requests.append(self.gsheets().request_cell_set_value(sheet_id, status_col, row_index, status       ))

                if status == 'issue_created':
                    jira_key  = jira_action.get('Jira_Key')
                    jira_link = '=HYPERLINK("https://jira.photobox.com/browse/{0}","{0}")'.format(jira_key)
                    requests.append(self.gsheets().request_cell_set_value(sheet_id, key_col      , row_index, jira_key ))
                    requests.append(self.gsheets().request_cell_set_value(sheet_id, jira_link_col, row_index, jira_link))





        return self.gsheets().execute_requests(self.file_id, requests)


from gw_bot.api.Slack_Commands_Helper import Slack_Commands_Helper
from osbot_gsuite.apis.commands.Slides_Commands            import Slides_Commands


class Lambda_Slides:
    def handle_lambda_event(self, event):

        params  = event.get('params')
        data    = event.get('data')
        if data is not None:
            channel = data.get('channel')
            team_id = data.get('team_id')
            return Slack_Commands_Helper(Slides_Commands).invoke(team_id, channel, params)


from osbot_gsuite.apis.GTypes import Merge_Type, Wrap_Strategy, RGB, Border_Style
from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Str import str_join


class GSheet:

    def __init__(self, gsheets, file_id, sheet_id, sheet_name):
        self.gsheets            = gsheets
        self.file_id            = file_id
        self.sheet_id           = sheet_id
        self.sheet_name         = sheet_name
        self.requests           = []
        self.requests_committed = []
        self.requests_results   = []

    def __repr__(self):
        return f"[GSheet] {self.sheet_name} (id: {self.sheet_id})"

    def add_request(self, request):
        self.requests.append(request)
        return self

    def add_request_borders(self, col_start, row_start, col_end=None, row_end=None, color=None, style=Border_Style.SOLID_MEDIUM):    # https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#UpdateBordersRequest
        color = color or RGB.BLACK                      # do this here to avoid the mutable warning when in ctor
        kwargs_range = {"col_start": col_start ,
                        "row_start": row_start ,
                        "col_end"  : col_end   ,
                        "row_end"  : row_end   }
        request = { "updateBorders": {"range" : self.format_range(**kwargs_range) ,
                                      "top"   : { "style": style  ,"color": color},
                                      "bottom": { "style": style  ,"color": color},
                                      "left"  : { "style": style  ,"color": color},
                                      "right" : { "style": style  ,"color": color}}}
                                      # innerHorizontal
                                      # innerVertical
        return self.add_request(request)

    def add_request_column_size(self, start_col, end_col, size):
        request = { "updateDimensionProperties": { "range"      : { "sheetId": self.sheet_id,"dimension": "COLUMNS",
                                                                  "startIndex": start_col,"endIndex": end_col },
                                                   "properties" : { "pixelSize": size},
                                                   "fields"     : "pixelSize" }}
        return self.add_request(request)

    def add_request_row_size(self, start_row, end_row, size):
        request = { "updateDimensionProperties": { "range"      : { "sheetId": self.sheet_id,"dimension": "ROWS",
                                                                  "startIndex": start_row,"endIndex": end_row },
                                                   "properties" : { "pixelSize": size},
                                                   "fields"     : "pixelSize" }}
        return self.add_request(request)

    def add_request_insert_column(self, start_col, end_col=None):
        start_index = start_col
        end_index   = end_col or start_col + 1
        request = { "insertDimension": { "range"      : { "sheetId": self.sheet_id,"dimension": "COLUMNS",
                                                          "startIndex": start_index,"endIndex": end_index },
                                        "inheritFromBefore": False }}
        return self.add_request(request)

    def add_request_insert_row(self, start_row, end_row=None):
        start_index = start_row
        end_index   = end_row or start_row + 1
        request = { "insertDimension": { "range"      : { "sheetId": self.sheet_id,"dimension": "ROWS",
                                                          "startIndex": start_index,"endIndex": end_index },
                                        "inheritFromBefore": False }}
        return self.add_request(request)

    def add_request_merge_cells(self, col_start, row_start, col_end=None, row_end=None, merge_type=Merge_Type.MERGE_ALL):
        kwargs_range = {"col_start": col_start ,
                        "row_start": row_start ,
                        "col_end"  : col_end   ,
                        "row_end"  : row_end   }
        request = { "mergeCells": {"range" : self.format_range(**kwargs_range) ,
                                      "mergeType"   : merge_type}}
        return self.add_request(request)

    def add_request_wrap_stategy(self,col_start, row_start, col_end=None, row_end=None, wrap_stategy=Wrap_Strategy.WRAP):
        kwargs_range = {"col_start": col_start ,
                        "row_start": row_start ,
                        "col_end"  : col_end   ,
                        "row_end"  : row_end   }
        request = { "updateCells"  : { "range": self.format_range(**kwargs_range) ,
                                       "rows": [{ "values": [ { "userEnteredFormat": { "wrapStrategy": wrap_stategy}}]}],
                    "fields"       :   "userEnteredFormat.wrapStrategy" }}
        request = {
			  "updateCells": {
				"range": {
				  "sheetId": self.sheet_id,
				  "startRowIndex": 0,
				  "startColumnIndex": 0,
                  "endRowIndex": 1,
                    "endColumnIndex": 1,
				},
				"rows": [
				  {
					"values": [
					  {
                          'userEnteredValue': {'stringValue': "value"},
						"userEnteredFormat": {
						  "wrapStrategy": wrap_stategy
						}
					  }
					]
				  }
				],
				"fields": "userEnteredValue.stringValue, userEnteredFormat.wrapStrategy"
			  }
			}


        pprint(request)
        self.add_request(request)
        return self
    def format_range(self, col_start, row_start, col_end=None, row_end=None):
        return { "sheetId"         : self.sheet_id              ,
                 "startRowIndex"   : row_start                  ,
                 "endRowIndex"     : row_end  or row_start + 1  ,  # default to only impacting the start_row_index
                 "startColumnIndex": col_start                  ,
                 "endColumnIndex"  : col_end  or col_start + 1  }  # default to only impacting the start_column_index

    def format_cell(self, background_color_rgb=None, foreground_color_rgb=None, text_align_horizontal=None, bold=False, font_size=None,wrap_strategy=None):
        textFormat      = {}
        cell            = {"textFormat": textFormat }
        fields        = []

        if background_color_rgb:
            cell['backgroundColor'] = background_color_rgb
            fields.append("backgroundColor")

        if bold is not None:
            textFormat['bold'] = bold
            fields.append("textFormat.bold")

        if text_align_horizontal:
            cell['horizontalAlignment'] = text_align_horizontal
            fields.append("horizontalAlignment")

        if font_size:
            textFormat['fontSize'] = font_size
            fields.append("textFormat.fontSize")

        if foreground_color_rgb:
            textFormat['foregroundColor'] = foreground_color_rgb
            fields.append("textFormat.foregroundColor")

        if wrap_strategy:
            cell['wrapStrategy'] = wrap_strategy
            fields.append("wrapStrategy")

        fields_format = f"userEnteredFormat({str_join(',',fields)})"
        cells_format  = {  "userEnteredFormat": cell }

        return cells_format, fields_format


    def request_format_cell(self, col_start, row_start, col_end=None, row_end=None, cell_color=None, text_color=None, text_align=None, bold=None, font_size=None, wrap_strategy=None):
        kwargs_range = {"col_start": col_start ,
                        "row_start": row_start ,
                        "col_end"  : col_end   ,
                        "row_end"  : row_end   }
        kwargs_cell  = {"background_color_rgb" : cell_color            ,
                        "foreground_color_rgb" : text_color            ,
                        "text_align_horizontal": text_align            ,
                        "bold"                 : bold                  ,
                        "font_size"            : font_size             ,
                        "wrap_strategy"        : wrap_strategy         }
        range        = self.format_range(**kwargs_range)
        cell, fields = self.format_cell(**kwargs_cell)
        return { "repeatCell": { "range" : range   ,
                                 "cell"  : cell    ,
                                 "fields": fields}},

    def bold(self, col_start, row_start, col_end=None, row_end=None, bold=True):
        kwargs_range = {"col_start": col_start ,
                        "row_start": row_start ,
                        "col_end"  : col_end   ,
                        "row_end"  : row_end   ,
                        "bold"     : bold      }
        request = self.request_format_cell(**kwargs_range)
        self.requests.append(request)
        return self

    def font_size(self, col_start, row_start, col_end=None, row_end=None, font_size=None):
        kwargs_range = {"col_start" : col_start      ,
                        "row_start" : row_start      ,
                        "col_end"   : col_end        ,
                        "row_end"   : row_end        ,
                        "font_size" : font_size      }
        request = self.request_format_cell(**kwargs_range)
        self.requests.append(request)
        return self

    def cell_color(self, col_start, row_start, col_end=None, row_end=None, cell_color=None):
        kwargs_range = {"col_start" : col_start      ,
                        "row_start" : row_start      ,
                        "col_end"   : col_end        ,
                        "row_end"   : row_end        ,
                        "cell_color" : cell_color      }
        request = self.request_format_cell(**kwargs_range)
        self.requests.append(request)
        return self

    def wrap_cells(self, col_start, row_start, col_end=None, row_end=None, wrap_strategy=Wrap_Strategy.WRAP):
        kwargs_range = {"col_start"     : col_start      ,
                        "row_start"     : row_start      ,
                        "col_end"       : col_end        ,
                        "row_end"       : row_end        ,
                        "wrap_strategy" : wrap_strategy      }
        request = self.request_format_cell(**kwargs_range)
        self.requests.append(request)
        return self

    def hide_grid_lines(self, value=True):
        request = {"updateSheetProperties": {"properties": {"sheetId": self.sheet_id,
                                                            "gridProperties": {"hideGridlines":value}},
                                             "fields": "gridProperties.hideGridlines"}}
        self.requests.append(request)
        return self

    def text_align(self, col_start, row_start, col_end=None, row_end=None, text_align=None):
        kwargs_range = {"col_start"  : col_start      ,
                        "row_start"  : row_start      ,
                        "col_end"    : col_end        ,
                        "row_end"    : row_end        ,
                        "text_align" : text_align      }
        request = self.request_format_cell(**kwargs_range)
        self.requests.append(request)
        return self

    def text_color(self, col_start, row_start, col_end=None, row_end=None, text_color=None):
        kwargs_range = {"col_start"  : col_start      ,
                        "row_start"  : row_start      ,
                        "col_end"    : col_end        ,
                        "row_end"    : row_end        ,
                        "text_color" : text_color      }
        request = self.request_format_cell(**kwargs_range)
        self.requests.append(request)
        return self

    def clear_formatting(self):
        self.requests.append(self.gsheets.request_clear_formatting(sheet_id=self.sheet_id))
        return self

    def set_value(self, col, row, value):
        request = self.gsheets.request_cell_set_value(sheet_id=self.sheet_id, col=col,row=row,  value=value)
        self.requests.append(request)
        return self

    def set_values(self, cells):
        for cell in cells:
            (col, row, value) = cell
            print(cell)
            self.set_value(col, row, value)
        return self

    def info(self):
        return { "file_id"              : self.file_id                ,
                 "sheet_id"             : self.sheet_id               ,
                 "sheet_name"           : self.sheet_name             ,
                 "# requests queued"    : len(self.requests          ),
                 "# requests committed" : len(self.requests_committed),
                 }

    def commit(self):
        if self.requests:
            self.gsheets.execute_requests(file_id=self.file_id, requests=self.requests)
            self.requests_committed.extend(self.requests)
            self.requests = []
            return True
        return False


from osbot_gsuite.apis.GSuite import GSuite
from osbot_utils.decorators.methods.cache_on_tmp import cache_on_tmp
from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Misc import list_set


class GPeople:

    def __init__(self):
        #self.api = GSuite().api('people')
        #self.api_v1 = GSuite().api('people_v1')
        pass

    def people(self):
        return GSuite().people_v1().people()

    def person(self, person_id):
        person_raw = self.person_raw(person_id)
        assert list_set(person_raw.keys()) == ['emailAddresses', 'etag', 'names', 'resourceName']
        email_addresses = person_raw.get('emailAddresses')
        names           = person_raw.get('names')
        #pprint(len(email_addresses))
        email_address   = self.primary_email_address(person_raw)
        display_name    = self.primary_display_name(person_raw)

        return {'email_address': email_address, 'display_name': display_name, 'person_id': person_id}

    def primary_display_name(self, person_raw):
        for name in person_raw.get('names'):
            if name.get('metadata').get('primary'):
                return name.get('displayName')

    def primary_email_address(self, person_raw):
        for email_address in person_raw.get('emailAddresses'):
            if email_address.get('metadata').get('primary'):
                return email_address.get('value')

    @cache_on_tmp()
    def person_raw(self, person_id, person_fields=None):
        if person_fields is None:
            person_fields = 'names,emailAddresses'
        elif person_fields == 'all':
            fields_list = [
                'addresses', 'ageRanges', 'biographies', 'birthdays', 'calendarUrls', 'clientData', 'coverPhotos',
                'emailAddresses', 'events', 'externalIds', 'genders', 'imClients', 'interests', 'locales', 'locations',
                'memberships', 'metadata', 'miscKeywords', 'names', 'nicknames', 'occupations', 'organizations',
                'phoneNumbers', 'photos', 'relations', 'sipAddresses', 'skills', 'urls', 'userDefined'
            ]
            person_fields = ','.join(fields_list)

        resource_name = f'people/{person_id}'
        return self.people().get(resourceName=resource_name, personFields=person_fields).execute()


import base64

from osbot_aws.helpers.Lambda_Helpers import slack_message
from osbot_aws.apis.Secrets import Secrets

from osbot_gsuite import version_osbot_gsuite
from osbot_gsuite.apis.GDrive import GDrive
from osbot_gsuite.apis.GSlides import GSlides
from osbot_utils.utils.Files import Files


class GDocs_Commands:
    gsuite_secret_id = 'gwbot_gsuite_token_2' #'gsuite_gsbot_user'

    @staticmethod
    def _gdrive():
        return GDrive(GDocs_Commands.gsuite_secret_id)

    @staticmethod
    def _gslides():
        return GSlides(GDocs_Commands.gsuite_secret_id)

    # @staticmethod
    # def _resolve_secret_id(team_id):
    #     if team_id == 'T7F3AUXGV':    return 'slack-gs-bot'
    #     if team_id == 'T0SDK1RA8':    return 'slack-gsbot-for-pbx'

    def _send_pdf_to_slack(pdf_data, channel): #todo: refactor to separate method
        import requests
        aws_secrets_id  = 'slack-bot-oauth'
        bot_token       = Secrets(aws_secrets_id).value()
        tmp_file        = Files.temp_file('.pdf')
        with open(tmp_file, "wb") as fh:
            fh.write(base64.decodebytes(pdf_data.encode()))

        my_file = {
            'file': ('/tmp/file.pdf', open(tmp_file, 'rb'), 'pdf')
        }
        payload = {
            "filename": '{0}.png'.format('pdf'),
            "token": bot_token,
            "channels": [channel],
        }
        requests.post("https://slack.com/api/files.upload", params=payload, files=my_file)
        #return my_file

    # @staticmethod
    # def list(team_id, channel, params):
    #     folder_id       = '16yOkKyi0TfOy3w4IMW40vo-pr--Wa1Y9'
    #     link_template   = 'https://drive.google.com/open?id='
    #     files           = GDocs_Commands._gdrive().files_in_folder(folder_id)
    #     attachment_text = ""
    #     text            = ":point_right: Found {0} files in the current gsbot docs folder".format(len(files))
    #     for file in files:
    #         title   = file.get('name')
    #         file_id = file.get('id')
    #         url   = link_template + file_id
    #         attachment_text += " • <{0}|{1}> \n".format(url,title)
    #
    #     attachments = [{ 'color':'good', 'text': attachment_text}]
    #
    #     return text, attachments

    @staticmethod
    def pdf(team_id=None, channel=None, params=None):
        if params and len(params) > 0:
            file_id = params.pop()
            slack_message(":point_right: creating pdf for file: `{0}`".format(file_id), [], channel, team_id)
            pdf_bytes = GDocs_Commands._gdrive().file_export(file_id)
            pdf_data = base64.b64encode(pdf_bytes).decode()
            GDocs_Commands._send_pdf_to_slack(pdf_data, channel)
            # payload = {
            #     'pdf_data'      : pdf_data                                  ,
            #     'title'         : 'export_pdf'                              ,
            #     'aws_secrets_id': GDocs_Commands._resolve_secret_id(team_id),
            #     'channel'       : channel
            # }
            #return len(pdf_data)

            #Lambda('utils.pdf_to_slack').invoke(payload)

        return None,None

    @staticmethod
    def version(team_id=None, channel=None, params=None):
        return version_osbot_gsuite,[]


from gsuite.GSuite import GSuite


class GSuite_Admin:

    def __init__(self):
        self.service = GSuite().admin_reports_v1()


    def get_api_data_raw(self, application, max_results=1000, page_token=None, start_time = None):
        return self.service.activities().list(userKey         = 'all'       ,
                                              applicationName = application ,
                                              maxResults      = max_results ,
                                              pageToken       = page_token  ,
                                              startTime       = start_time  ).execute()

    def get_api_data(self,  application, number_of_items_to_fetch = 100, max_results = 100):                # this doesn't scale for the calendar one since it has tons on entries
        all_items = []
        next_page_token = None
        if number_of_items_to_fetch > 1000:
            number_of_items_to_fetch = 1000         # this is limitation of gsuite logs

        def get_results(page_token):
            results = self.get_api_data_raw(application, number_of_items_to_fetch, page_token)
            all_items.extend(results['items'])
            return results.get('nextPageToken')

        next_page_token = get_results(next_page_token)

        while next_page_token is not None:
            if max_results is not None:
                if len (all_items) >= max_results:
                    break;
            next_page_token = get_results(next_page_token)

        return all_items


# from osbot_aws.helpers.Lambda_Package import Lambda_Package
# from osbot_utils.utils.Files import Files
#
#
# class Deploy:
#
#     def __init__(self, lambda_name):
#         self.package         = Lambda_Package(lambda_name)
#         self.tmp_s3_bucket = 'gs-lambda-tests'
#         self.tmp_s3_key    = 'gsbot/{0}.zip'.format(lambda_name)
#         self.setup()
#
#     def setup(self):
#         (self.package._lambda.set_s3_bucket(self.tmp_s3_bucket)
#                              .set_s3_key   (self.tmp_s3_key)
#                              .set_xrays_on())
#
#
#
#     def deploy(self, delete_before=False):
#
#         if delete_before:
#             self.package.delete()
#         code_folder = Files.path_combine(__file__,'..')
#         self.package.add_folder(code_folder)
#         self.package.add_root_folder()
#         self.package.add_osbot_utils()
#         return self.package.update()





from osbot_gsuite.apis.create_slides.GSBot_Helpers       import GSBot_Helper
from osbot_gsuite.apis.create_slides.GSBot_to_GDrive     import GSBot_to_GDrive
from osbot_gsuite.apis.create_slides.Slides_for_Projects import Slides_for_Projects
from osbot_utils.utils.Dev import Dev
from osbot_utils.utils import Misc


class Project_Slides:
    def __init__(self):
        self.file_id      = None
        self.file_type    = None
        self.jira_key     = None
        self.title        = None
        self._project     = None
        self._slides_ids  = None
        self.gsuite_secret_id    = 'gsuite_gsbot_user'
        self.jira_link           = 'https://jira.photobox.com/browse/{0}'
        self.slides_for_projects = Slides_for_Projects(self.gsuite_secret_id)

    def gdrive (self): return self.slides_for_projects.gdrive()
    def gslides(self): return self.slides_for_projects.gslides()

    def project(self):
        if self._project is None:
            self._project = self.slides_for_projects.gs_project_by_Id(self.jira_key)
            if self._project:
                # remove fields that we don't need to see in the main slides
                for field in ['Brands','Components','Created','Creator', 'Epic Link',
                              'Impacts','Latest_Information','Likelihood','Parent',
                              'Project', 'Reporter', 'Risk Owner','Updated',
                              'ISO27001 Standard','GDPR Article']:
                    del self._project[field]

                # fix description field
                if self._project.get('Description') is None:
                    self._project['Description'] = 'No description, please add one one using the link below'
                self._project['Description'] = Misc.remove_html_tags(self._project['Description'])

        return self._project

    def fix_slide_id(self,text):
        return text.replace(' ', '_').replace('(', '_').replace(')', '_')

    def slides_ids(self):
        if self._slides_ids is None:
            self._slides_ids = list(set(self.gslides().slides_indexed_by_id(self.file_id)))
        return self._slides_ids

    def delete_presentation(self,jira_key=None):
        if jira_key:
            self.resolve_file_id(jira_key)
        self.gdrive().file_delete(self.file_id)
        return self

    def delete_slide(self, slide_id):
        if slide_id in self.slides_ids():
            self.gslides().slide_delete(self.file_id, slide_id)
            self._slides_ids.remove(slide_id)
        return self

    def delete_all_content_slides(self):
        for slide_id in list(self.slides_ids()):
            if slide_id != 'slide_title':
                self.delete_slide(slide_id)
        return self

    def resolve_file_id(self, jira_key):
        self.jira_key = jira_key
        project = self.project()
        if project:
            summary = project.get("Summary")
            self.title = "GS Project - {0} - {1}".format(jira_key, summary)

            file = self.gdrive().find_by_name(self.title)
            if file is None:
                template_file_id = '1oApT1ujAcPuw_zMk8O5fVew9Mo4xNA28VHdLoq5DfX0'
                parent_folder    = '16yOkKyi0TfOy3w4IMW40vo-pr--Wa1Y9'
                self.file_id = self.gslides().presentation_copy(template_file_id, self.title, parent_folder)
            else:
                self.file_id = file.get('id')
        return self

    def set_jira_key(self, new_key):
        self._project = None       # force data reload
        self.jira_key = new_key
        return self
    # slides 'add helpers'

    def create_slide(self, requests, slide_id, title, x=10, y=10, width=500, height=50, size=26):
        slide_id = self.fix_slide_id(slide_id)
        self.delete_slide(slide_id)
        title_id = '{0}_title'.format(slide_id)
        requests.extend([ self.gslides().slide_create_request                      (slide_id,                             ),
                          self.gslides().element_create_shape_request              (slide_id, x,y,width, height, title_id  ),
                          self.gslides().element_insert_text_request               (title_id,title                        ),
                          self.gslides().element_set_text_style_requests__for_title(title_id, size                         )])
        return self

    def add_text(self, requests, slide_id, text, x, y, width, height,size=None, bullets= False):
        text_id = Misc.random_string_and_numbers(6, 'text_')
        requests.extend([self.gslides().element_create_shape_request              (slide_id, x, y, width, height, text_id ),
                         self.gslides().element_insert_text_request               (text_id ,text                         )])
        if size:
            requests.append({'updateTextStyle': {'objectId': text_id,
                                                 'style': {'fontSize': {'magnitude': size, 'unit': 'PT'}},
                                                 'fields': 'fontSize'}})


        if bullets:
            requests.append({"createParagraphBullets": { "objectId": text_id,
                                                         "bulletPreset": "BULLET_ARROW_DIAMOND_DISC",
                                                         "textRange": { "type": "ALL" }}})

        return self

    def add_text_with_keys_titles(self, requests, slide_id, keys, x, y, width, height, size=None):
        if not keys: return self

        issues = list(self.slides_for_projects.elastic().search_on_field_for_values("Key.keyword", keys))
        text = ''
        for issue in issues:
            text += "{0} \n".format(issue.get('Summary'))
        text = text.strip()
        return self.add_text(requests, slide_id, text, x, y, width, height, size, bullets=True)


    def add_link(self, requests, slide_id, text, link, x, y, width, height, size=None):
        text_id = Misc.random_string_and_numbers(6, 'text_')
        requests.extend([self.gslides().element_create_shape_request              (slide_id, x, y, width, height, text_id ),
                         self.gslides().element_insert_text_request               (text_id ,text                          ),
                         self.gslides().element_set_text_style_requests           (text_id, {'link': {'url': link }},'link'   )])
        if size:
            requests.append({'updateTextStyle': {'objectId': text_id,
                                                 'style': {'fontSize': {'magnitude': size, 'unit': 'PT'}},
                                                 'fields': 'fontSize'}})

        return self

    def add_image_from_expanded_graph(self, requests, slide_id, graph_name, depth, links_path):
        png_file  = GSBot_Helper().get_png_from_expanded_graph(graph_name, depth,links_path)
        image_id  = GSBot_to_GDrive(self.gsuite_secret_id).upload_png_file_to_gdrive(png_file)
        image_url = "https://lh3.google.com/u/1/d/{0}".format(image_id)
        requests.append(self.gslides().element_create_image_request(slide_id, image_url, 10, 70, 600, 300))

    def add_image_from_new_graph(self, requests, slide_id, start, direction, depth, view):
        png_file  = GSBot_Helper().get_png_from_new_graph(start, direction, depth, view)
        image_id  = GSBot_to_GDrive(self.gsuite_secret_id).upload_png_file_to_gdrive(png_file)
        image_url = "https://lh3.google.com/u/1/d/{0}".format(image_id)
        requests.append(self.gslides().element_create_image_request(slide_id, image_url, 10, 70, 600, 300))

    def add_image_from_elk_dashboard(self, requests, slide_id, jira_key,x, y, width,height):
        png_file  = GSBot_Helper().get_png_from_elk_dashboard(jira_key)
        image_id  = GSBot_to_GDrive(self.gsuite_secret_id).upload_png_file_to_gdrive(png_file)
        image_url = "https://lh3.google.com/u/1/d/{0}".format(image_id)
        requests.append(self.gslides().element_create_image_request(slide_id, image_url, x, y, width,height))
        return self

    def add_image_from_browser_risks(self, requests, slide_id, risks,x, y, width,height):
        png_file  = GSBot_Helper().get_png_from_browser_risks(risks)
        image_id  = GSBot_to_GDrive(self.gsuite_secret_id).upload_png_file_to_gdrive(png_file)
        image_url = "https://lh3.google.com/u/1/d/{0}".format(image_id)
        requests.append(self.gslides().element_create_image_request(slide_id, image_url, x, y, width,height))
        return self


    def add_new_graph(self, title, start, direction, depth, view):
        if self.project():
            slide_id  = 'slide_' + title.replace(' ', '_').replace('(', '_').replace(')', '_')
            jira_link = self.jira_link.format(start)
            requests = []
            (
                self.create_slide                 (requests, slide_id, title                         ),
                self.add_image_from_new_graph     (requests, slide_id, start, direction, depth, view ),
                self.add_link                     (requests, slide_id, start, jira_link, 630, 330, 700, 50)
            )
            self.gslides().execute_requests(self.file_id, requests)
        return self

    def add_expanded_graph(self, title, graph_name, depth, links_path):
        if self.project():
            slide_id        = 'slide_' + title.replace(' ','_').replace('(','_').replace(')','_')
            requests = []
            (
                self.create_slide                 (requests, slide_id, title                        )
                    .add_image_from_expanded_graph(requests, slide_id, graph_name, depth, links_path)
            )
            self.gslides().execute_requests(self.file_id, requests)
        return self

    def add_project_1_pager(self, jira_key,with_images=True):
        self.set_jira_key(jira_key)
        if self.project():
            issue_data      = self.project()
            slide_id        = 'slide_project_{0}'.format(jira_key)
            title           = issue_data.get('Summary')
            jira_link       = self.jira_link.format(jira_key)
            description     = Misc.remove_html_tags(issue_data.get('Description'))
            fixes           = issue_data.get('Issue Links').get('fixes')
            if not fixes:
                fixes       = issue_data.get('Issue Links').get('reduces risk of')  # for now add these if no 'fixes' exists
            text_why        = 'Why are we doing this project (i.e. what does it fix):'
            #Dev.pprint(self.project())
            r2s = GSBot_Helper().find_R2s_for_jira_issue(jira_key)
            r2s_issues = list(self.slides_for_projects.elastic().search_on_field_for_values("Key.keyword", r2s))
            risks = []
            for r2_issue in r2s_issues:
                risks.append("r{0}:2".format(r2_issue.get('Summary').split('-')[0].replace('.','_').strip()))
            risks = ','.join(risks)
            Dev.pprint(risks)
            requests = []

            (
                self.create_slide                 (requests, slide_id, title         , size =20)
                    .add_link                     (requests, slide_id, self.jira_key, jira_link,  390, 375, 700,  60 ,6)
                    .add_text                     (requests, slide_id, description   , 10,  70, 400,100, 9)
                    .add_text                     (requests, slide_id, text_why      , 10, 210, 400,100, 14)
                    .add_text_with_keys_titles    (requests, slide_id, fixes         , 10, 240, 400, 100, 9)
            )
            self.gslides().execute_requests(self.file_id, requests)

            if with_images:
                requests = []
                self.add_image_from_elk_dashboard (requests, slide_id, jira_key      , 450,  -40, 260, 360)
                self.gslides().execute_requests(self.file_id, requests)

                requests = []
                self.add_image_from_browser_risks (requests, slide_id, risks         , 450,  230 , 260, 210)
                self.gslides().execute_requests(self.file_id, requests)
            
        return self

        
    # Creating Slide's content

    def set_content_title_slide(self, slide_type=None):
        if self.project():
            if slide_type is None: slide_type = 'GS Project'
            title    = self.project().get('Summary')
            subtitle = "{0} : {1}".format(slide_type,self.project().get('Key'))
            self.gslides().element_set_text(self.file_id, 'slide_title___title'    , title)
            self.gslides().element_set_text(self.file_id, 'slide_title___subtitle' , subtitle)
        return self

    def add_slide_with_title(self):
        data       = self.project()
        key        = data.get('Key')
        summary    = data.get('Summary')
        issue_type = data.get('Issue Type')
        jira_link = self.jira_link.format(key)
        requests   = []
        slide_id = self.fix_slide_id('slide_{0}'.format(summary))
        (
            self.create_slide(requests, slide_id, summary, 100,100,500,250,50)
                .add_text    (requests, slide_id, issue_type, 100,250,500,250)
                .add_link    (requests, slide_id, key, jira_link, 630, 330, 700, 50)
        )

        self.gslides().execute_requests(self.file_id, requests)
        return self

    def add_slide_project_details(self):
        if self.project():
            slide_id        = 'slide_project_details'
            title           = 'Project Details'.format(self.jira_key)
            description     = self.project().get('Description')
            jira_link       = self.jira_link.format(self.jira_key)
            requests        = []
            (
                self.create_slide(requests, slide_id, title                                         )
                    .add_text    (requests, slide_id, description,                10, 100, 650, 150 )
                    .add_link    (requests, slide_id, self.jira_key, jira_link,  630, 330, 700,  50 )
            )
            self.gslides().execute_requests(self.file_id, requests)
        return self

    def add_slide_with_table_with_key_details(self, name, keys):
        max_text_size = 450
        if keys is None: keys = []
        services = list(self.slides_for_projects.elastic().search_on_field_for_values("Key.keyword", keys))
        table_data = [['Jira Id', 'Title', 'Description' ]]

        for service in services:
            jira_key    = service.get('Key')
            title       = service.get('Summary')
            description = service.get('Description')
            if description is None: description = 'No Description, please add one in Jira'
            description = Misc.remove_html_tags(description).replace('\n',' ')
            description = (description[:max_text_size] +  ' ...') if len(description) > max_text_size else description
            table_data.append([jira_key,title,description])
        slide_id = 'slide_{0}'.format(name.lower()).replace(' ','_').replace('(','_').replace(')','_')
        title    = '{0}'.format(name)
        table_id = self.gslides().add_slide_with_table_from_array(self.file_id, slide_id, title, table_data,[50,100,545])

        requests = []
        for index, row in enumerate(table_data):
            jira_id = row.pop(0)
            link = self.jira_link.format(jira_id)
            requests.append(self.gslides().element_set_table_text_style_request(table_id, index +1 ,0,{'link': {'url': link }},'link'))

        self.gslides().execute_requests(self.file_id,requests)

    def add_slide_with_issues_linked(self, link_type, name, max_rows  = 6):
        keys = self.project().get('Issue Links').get(link_type)
        if keys:
            size      = len(keys)
            page      = 0
            max_pages = round( (size+1) /max_rows)
            for i in range(0, size, max_rows):
                page  +=1
                title = "{0} ({1} of {2})".format(name, page, max_pages)
                values = keys[i:i + max_rows]
                self.add_slide_with_table_with_key_details(title,values)
        return self

    def add_slide_services_used(self): return self.add_slide_with_issues_linked('uses'           , "Services Used")
    def add_slide_risks_reduced(self): return self.add_slide_with_issues_linked('reduces risk of', "Risks Reduced")
    def add_slide_delivers     (self): return self.add_slide_with_issues_linked('delivers'       , "Delivers")
    def add_slide_affects      (self): return self.add_slide_with_issues_linked('affects'        , "Affects"      )

    def add_slide_with_graph_risks_to_R1 (self): return self.add_expanded_graph  ('Graph - Risks (path to R1)', self.jira_key       , 5, 'risks_up'                        )
    def add_slide_with_graph_stakeholders(self): return self.add_expanded_graph  ('Graph - Stakeholders'      , self.jira_key       , 5, 'risks_up,affects,stakeholders_up')
    def add_slide_with_graph_all_1       (self): return self.add_new_graph       ('Graph (all nodes depth 1)' , self.jira_key, 'all', 1, 'colors'                          )

    def add_slide_with_project_1_pager   (self, project_id,with_images=True):
        try:
            return self.add_project_1_pager(project_id,with_images)
        except Exception as error:
            Dev.pprint("[add_slide_with_project_1_pager][error] {0}".format(error))
            return self


    def add_slides_with_project_1_pager  (self, project_ids):
        for project_id in project_ids:
            self.add_slide_with_project_1_pager(project_id)
        return self

    def add_metadata_slide(self):
        if self.project():
            del self._project['Description']
            slide_id = 'slide_metadata'
            title    = 'Project {0} Metadata'.format(self.jira_key)
            #self.gslides().slide_delete(self.file_id, slide_id)                         # this one needs to be done outside since it will fail on first creation
            self.gslides().add_slide_with_table_from_object(self.file_id, slide_id, title, self.project())
        return self


    def create_slides_text_content(self, jira_key):
        (
            self.resolve_file_id                 (jira_key)
                .delete_all_content_slides       ()
                .set_content_title_slide         ()
                #.add_slide_project_details       ()
                .add_slide_with_project_1_pager  (jira_key)
                .add_slide_services_used         ()
                .add_slide_risks_reduced         ()
                .add_slide_delivers()
                .add_slide_affects               ()
        )
        return self.gdrive().file_weblink(self.file_id)

    def create_slides_graphs(self):
        (
            self.add_slide_with_graph_risks_to_R1()
                .add_slide_with_graph_stakeholders()
                .add_slide_with_graph_all_1()
        )
        return self

    def create_slides_appendix(self):
        (
            self.add_metadata_slide()
        )
        return self



    # Function Slides generation (move to separate file

    def add_slide_function_details(self):
        if self.project():
            slide_id        = 'slide_function_details'
            title           = 'Function Details'.format(self.jira_key)
            description     = self.project().get('Description')
            jira_link       = self.jira_link.format(self.jira_key)
            requests        = []
            (
                self.create_slide(requests, slide_id, title                                         )
                    .add_text    (requests, slide_id, description,                10, 100, 650, 150 )
                    .add_link    (requests, slide_id, self.jira_key, jira_link,  630, 330, 700,  50 )
            )
            self.gslides().execute_requests(self.file_id, requests)

        return self

    def create_slides_for_function(self, key_function):
        (
            self.resolve_file_id             (key_function)
                .delete_all_content_slides   ()
                .set_content_title_slide     ("GS Function")
                .add_slide_function_details  ()
        )

        keys = self.project().get('Issue Links').get('is parent of')

        for key in keys:
            self.set_jira_key(key)
            title      = self.project().get('Summary')
            issue_type = self.project().get('Issue Type')
            self.add_slide_with_title()
            if issue_type == 'Programme':
                self.add_slide_with_issues_linked('delivered by', title) #"Projects")
            if issue_type == 'Business entity':
                self.add_slide_with_issues_linked('is parent of', title) #"Services")


        #         .set_jira_key                (key_projects)
        #         .add_slide_with_issues_linked('delivered by', "Projects")
        #         .set_jira_key                (key_services)
        #         .add_slide_with_issues_linked('is parent of', "Services")
        #
        #         #.add_new_graph          ('key_projects' , key_projects, 'all', 1, 'colors')
        #         #.add_new_graph          ('key_services' , key_services , 'all', 1, 'colors')
        # )
        return self.gdrive().file_weblink(self.file_id)


    def create_project_slides_for_functions(self, key_function,team_id, channel):
        (
            self.resolve_file_id             (key_function)
                .delete_all_content_slides   ()
                .set_content_title_slide     ("PROJECTS")
        )
        delivered_by = self.project().get('Issue Links').get('delivered by')
        delivered_by.remove('GSSP-327')        # for now manually remove this project (GSP BAU)

        #weblink = self.gdrive().file_weblink(self.file_id)
        #slack_message(':point_rigth: batch creating slides for {0}'.format(self.project))

        issues = list(self.slides_for_projects.elastic().search_on_field_for_values("Key.keyword", delivered_by))
        for issue in issues:
            try:
                if issue.get('Key') == 'GSSP-118': continue #skip fix since it has been done
                #if issue.get('Key') == 'GSSP-120': continue # detect
                #if issue.get('Key') == 'GSSP-121': continue # risk
                if issue.get('Key') == 'GSSP-113': continue # gs internal
                if issue.get('Key') == 'GSSP-115': continue # infosec
                self.set_jira_key(issue.get('Key'))
                if self.project():
                    self.add_slide_with_title()
                    delivered_by = self.project().get('Issue Links').get('delivered by')
                    for project in delivered_by:
                        self.add_slide_with_project_1_pager(project)
            except Exception as error:
                Dev.pprint('[create_project_slides_for_functions][error] {0}'.format(error))

        return self.gdrive().file_weblink(self.file_id)


from osbot_aws.Dependencies import load_dependency
from osbot_aws.helpers.Lambda_Helpers import log_to_elk


def run(event, context):
    try:
        load_dependency("gmail")
        load_dependency("elastic")
        load_dependency("slack")
        load_dependency('requests')

        from osbot_gsuite.apis.handlers.Lambda_GDocs import Lambda_GDocs
        return Lambda_GDocs().handle_lambda_event(event)                      # invoke lambda handler from Lambda_Slides class
        return "200 OK"

    except Exception as error:
        message = "[lambda_gdocs] Error: {0}".format(error)
        log_to_elk(message, level='error')
        return message